import unittest

import numpy as np

from ms_peak_picker.peak_set import FittedPeak, PeakSet, _has_c, _PeakSet
from ms_peak_picker.peak_index import PeakIndex
from ms_peak_picker.reprofile import reprofile
from ms_peak_picker.peak_picker import pick_peaks, PeakProcessor, _has_c as _has_peak_picker_c

from ms_peak_picker.test.common import make_peak


points = [(276.5, 2e4), (576.5, 8e4), (862.1, 15e4)]


class TestPeakSet(unittest.TestCase):
    peak_set_type = PeakSet

    def make_peaks(self):
        inst = self.peak_set_type([make_peak(*point) for point in points])
        inst.reindex()
        return inst

    def test_construct(self):
        inst = self.make_peaks()
        self.assertEqual(len(inst), 3)

    def test_has_peak(self):
        inst = self.make_peaks()
        self.assertIsNotNone(inst.has_peak(576.5))
        self.assertIsNone(inst.has_peak(1576.5))
        for peak in inst:
            assert inst.has_peak(peak.mz) is not None

    def test_between(self):
        inst = self.make_peaks()
        self.assertIn(inst.has_peak(576.5), list(inst.between(300.0, 1000.01)))
        self.assertNotIn(inst.has_peak(276.5), list(inst.between(300.0, 1000.01)))

    def test_clone(self):
        inst = self.make_peaks()
        dup = inst.clone()
        for a, b in zip(inst.peaks, dup.peaks):
            self.assertEqual(a, b)


if _has_c:
    class TestPythonPeakSet(TestPeakSet):
        peak_set_type = _PeakSet


class TestPeakPicker(unittest.TestCase):
    @staticmethod
    def make_profile():
        return reprofile([make_peak(*point) for point in points])

    def test_peak_processor(self):
        mzs, intensities = self.make_profile()
        proc = PeakProcessor()
        proc.discover_peaks(mzs, intensities)
        peaks = proc.peak_data
        assert len(peaks) == 3

    def test_python_impl_peak_processor(self):
        if _has_peak_picker_c:
            from ms_peak_picker.peak_picker import _PeakProcessor
            mzs, intensities = self.make_profile()
            proc = _PeakProcessor()
            proc.discover_peaks(mzs, intensities)
            peaks = proc.peak_data
            assert len(peaks) == 3

    def test_pick_peaks(self):
        mzs, intensities = self.make_profile()
        peaks = pick_peaks(mzs, intensities)
        peak = peaks.has_peak(276.5, 1e-5)
        self.assertIsNotNone(peak)


class TestPeakIndex(TestPeakSet):
    @staticmethod
    def make_peaks():
        inst = PeakSet([make_peak(*point) for point in points])
        inst.reindex()
        inst = PeakIndex(np.array([]), np.array([]), inst)
        return inst


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np

from ms_peak_picker.fticr_denoising import denoise
from ms_peak_picker.test.common import make_peak
from ms_peak_picker import reprofile, pick_peaks


class TestDenoising(unittest.TestCase):
    def make_data(self):
        peak = make_peak(200, 1e4)
        x = np.arange(0, 1000, 0.01)
        y = (np.random.random(x.size) + 1) * 100
        y[19700:20300] += reprofile([peak])[1]
        return x, y

    def test_denoise(self):
        x, y = self.make_data()
        denoised_x, denoised_y = denoise(x, y, window_size=2.0)
        assert y.mean() >= 100.0
        assert denoised_y.mean() < 1.0
        assert len(pick_peaks(denoised_x, denoised_y)) == 1

    def test_no_change(self):
        x, y = self.make_data()
        denoised_x, denoised_y = denoise(x, y, window_size=2.0, scale=0)
        assert np.allclose(x, denoised_x)
        assert np.allclose(y, denoised_y)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from ms_peak_picker import FittedPeak
from ms_peak_picker.peak_statistics import GaussianModel
from ms_peak_picker.peak_set import simple_peak, is_peak


class PeakSetReprofiler(object):

    def __init__(self, models, dx=0.01):
        self.models = sorted(models, key=lambda x: x.center)
        self.dx = dx
        self.gridx = None
        self.gridy = None
        self._build_grid()

    def _build_grid(self):
        lo = self.models[0].center
        hi = self.models[-1].center
        self.gridx = np.arange(max(lo - 3, 0), hi + 3, self.dx, dtype=np.float64)
        self.gridy = np.zeros_like(self.gridx, dtype=np.float64)

    def _reprofile(self):
        nmodels = len(self.models)
        for i, x in enumerate(self.gridx):
            y = 0
            if i % 1000 == 0:
                print(i, x)
            offset = self._find_starting_model(x)
            j = offset - 1
            while j > 0:
                model = self.models[j]
                if (x - model.center) > 3:
                    break
                pred = model.predict(x)
                y += pred
                j -= 1
            j = offset
            while j < nmodels:
                model = self.models[j]
                if (model.center - x) > 3:
                    break
                pred = model.predict(x)
                y += pred
                j += 1

            self.gridy[i] = y

    def _find_starting_model(self, x):
        lo = 0
        hi = len(self.models)
        while (lo != hi):
            mid = (hi + lo) // 2
            model = self.models[mid]
            center = model.peak.mz
            err = center - x
            if abs(err) < 0.0001:
                return mid
            elif (hi - lo) == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid

    def reprofile(self):
        self._reprofile()
        return self.gridx, self.gridy


def models_from_peak_sets(peak_sets, max_fwhm=0.2, model_cls=None, default_fwhm=0.1):
    if model_cls is None:
        model_cls = GaussianModel
    models = []
    for peaks in peak_sets:
        for peak in peaks:
            try:
                if peak.full_width_at_half_max > max_fwhm:
                    continue
            except AttributeError:
                peak = simple_peak(peak.mz, peak.intensity, default_fwhm)
            models.append(model_cls(peak))
    return models


def reprofile(peaks, max_fwhm=0.2, dx=0.01, model_cls=GaussianModel, default_fwhm=0.1):
    """Converts fitted peak centroids into theoretical profiles derived from
    its fitted parameters and a theoretical shape model.

    Parameters
    ----------
    peaks : Iterable of FittedPeak or Iterable of Iterable of FittedPeak
        The peaks to convert back into profiles. If a list of peaks is
        provided, those peaks will be converted into profiles. If a list
        of lists of peaks is provided, all peaks will be summed when
        reconstructing the total signal at a given coordinate, and an
        average will be taken.
    max_fwhm : float, optional
        The maximum full width at half max to consider when selecting
        peaks to contribute to the modeled profiles.
    dx : float, optional
        The spacing of the m/z grid to use.
    model_cls : type, optional
        A type descending from :class:`ms_peak_picker.peak_statistics.PeakShapeModel`.
        Defaults to :class:`ms_peak_picker.peak_statistics.GaussianModel`.

    Returns
    -------
    mz_array: np.ndarray[float64]
        The m/z grid reconstructed from the fitted peaks
    intensity_array: np.ndarray[float64]
        The modeled total signal at each grid point
    """
    if is_peak(peaks[0]):
        peaks = [peaks]
    models = models_from_peak_sets(peaks, max_fwhm, model_cls, default_fwhm)
    task = PeakSetReprofiler(models, dx)
    x, y = task.reprofile()
    y /= len(peaks)
    return x, y


try:
    from ms_peak_picker._c.peak_statistics import PeakSetReprofiler
except ImportError:
    pass



# cython: embedsignature=True

cimport cython
from cython cimport parallel
cimport numpy as np
from libc cimport math
from libc.stdlib cimport malloc, calloc, free
from multiprocessing import cpu_count
import numpy as np

from cpython cimport PyFloat_AsDouble
from cpython cimport PyInt_AsLong
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM

np.import_array()

cdef long num_processors = PyInt_AsLong(cpu_count())


@cython.cdivision(True)
@cython.boundscheck(False)
cdef size_t binsearch(double* array, double x, size_t n) nogil:
    cdef:
        size_t lo, hi, mid
        double y, err

    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        y = array[mid]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


cdef struct spectrum_holder:
    double* mz
    double* intensity
    size_t size


cdef int prepare_arrays(list arrays, spectrum_holder** out):
    cdef:
        size_t i, n
        spectrum_holder* converted
        np.ndarray[double, ndim=1, mode='c'] mz
        np.ndarray[double, ndim=1, mode='c'] inten

    n = len(arrays)
    converted = <spectrum_holder*>malloc(sizeof(spectrum_holder) * n)
    if converted == NULL:
        return 1
    for i in range(n):
        mz, inten = arrays[i]
        converted[i].mz = &mz[0]
        converted[i].intensity = &inten[0]
        converted[i].size = mz.shape[0]
    out[0] = converted
    return 0


@cython.cdivision(True)
@cython.boundscheck(False)
cpdef average_signal(object arrays, double dx=0.01, object weights=None):
    """Average multiple spectras' intensity arrays, with a common m/z axis

    Parameters
    ----------
    arrays : :class:`list` of pairs of :class:`np.ndarray`
        The m/z and intensity arrays to combine
    dx : float, optional
        The m/z resolution to build the averaged m/z axis with
    weights : :class:`list` of :class:`float`, optional
        Weight of each entry in `arrays`. Defaults to 1.0 for each if not provided.

    Returns
    -------
    mz_array: :class:`np.ndarray`
    intensity_array: :class:`np.ndarray`
    """
    cdef:
        double lo, hi
        double mz_j, mz_j1
        double inten_j, inten_j1
        double x, contrib
        double scan_weight
        size_t i, j, n, n_arrays, n_points
        long k_array, n_scans, n_workers, worker_block
        object omz, ointen
        list convert

        spectrum_holder* spectrum_pairs
        spectrum_holder pair

        np.ndarray[double, ndim=1, mode='c'] mz_array
        np.ndarray[double, ndim=1, mode='c'] intensity_array
        np.ndarray[double, ndim=1, mode='c'] mz
        np.ndarray[double, ndim=1, mode='c'] inten
        double* pweights
        double* pmz
        double* pinten
        double* intensity_array_local

    if weights is None:
        weights = [1 for omz in arrays]
    elif len(arrays) != len(weights):
        raise ValueError("`arrays` and `weights` must have the same length")

    convert = []
    for omz, ointen in arrays:
        mz = omz.astype(np.double)
        inten = ointen.astype(np.double)
        convert.append((mz, inten))

    lo = convert[0][0][0]
    hi = 0
    for mz, inten in convert:
        lo = min(mz[0], lo)
        hi = max(mz[mz.shape[0] - 1], hi)

    lo = max(lo - 1, 0)
    hi += 1

    n_scans = len(convert)

    pweights = <double*>malloc(sizeof(double) * n_scans)
    for i in range(n_scans):
        pweights[i] = PyFloat_AsDouble(float(weights[i]))

    prepare_arrays(convert, &spectrum_pairs)

    mz_array = np.arange(lo, hi, dx)
    intensity_array = np.zeros_like(mz_array)
    n_points = mz_array.shape[0]
    with nogil:
        if n_scans < num_processors:
            n_workers = n_scans
        else:
            n_workers = num_processors
        for k_array in parallel.prange(n_workers):
            intensity_array_local = <double*>calloc(sizeof(double), n_points)
            pair = spectrum_pairs[k_array]
            pmz = pair.mz
            pinten = pair.intensity
            contrib = 0
            scan_weight = pweights[k_array]
            for i in range(n_points):
                x = mz_array[i]
                j = binsearch(pmz, x, pair.size)
                mz_j = pmz[j]
                if mz_j < x and j + 1 < pair.size:
                    mz_j1 = pmz[j + 1]
                    inten_j = pinten[j]
                    inten_j1 = pinten[j + 1]
                elif mz_j > x and j > 0:
                    mz_j1 = mz_j
                    inten_j1 = pinten[j]
                    mz_j = pmz[j - 1]
                    inten_j = pmz[j - 1]
                else:
                    continue

                contrib = ((inten_j * (mz_j1 - x)) + (inten_j1 * (x - mz_j))) / (mz_j1 - mz_j)
                intensity_array_local[i] += contrib * scan_weight
            for i in range(n_points):
                intensity_array[i] = intensity_array[i] + intensity_array_local[i]
            free(intensity_array_local)
        scan_weight = 0
        for i in range(n_scans):
            scan_weight += pweights[i]
        for i in range(n_points):
            intensity_array[i] /= scan_weight
        free(spectrum_pairs)
        free(pweights)
    return mz_array, intensity_array



import math
from collections import defaultdict

import numpy as np


def _binsearch(array, x):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        y = array[mid]
        err = y - x
        if hi - lo == 1:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


def average_signal(arrays, dx=0.01, weights=None):
    """Average multiple spectras' intensity arrays, with a common m/z axis

    Parameters
    ----------
    arrays : :class:`list` of pairs of :class:`np.ndarray`
        The m/z and intensity arrays to combine
    dx : float, optional
        The m/z resolution to build the averaged m/z axis with
    weights : :class:`list` of :class:`float`, optional
        Weight of each entry in `arrays`. Defaults to 1.0 for each if not provided.

    Returns
    -------
    mz_array: :class:`np.ndarray`
    intensity_array: :class:`np.ndarray`
    """
    if weights is None:
        weights = [1 for omz in arrays]
    elif len(arrays) != len(weights):
        raise ValueError("`arrays` and `weights` must have the same length")
    lo = max(min([x.min() for x, y in arrays]) - 1, 0)
    hi = max([x.max() for x, y in arrays]) + 1
    arrays = [(x.astype(float), y.astype(float)) for x, y in arrays]
    if isinstance(dx, float):
        mz_array = np.arange(lo, hi, dx)
    elif isinstance(dx, np.ndarray):
        mz_array = dx
    intensity_array = np.zeros_like(mz_array)
    arrays_k = 0
    for mz, inten in arrays:
        weight = weights[arrays_k]
        arrays_k += 1
        contrib = 0
        for i, x in enumerate(mz_array):
            j = _binsearch(mz, x)
            mz_j = mz[j]
            if mz_j < x and j + 1 < mz.shape[0]:
                mz_j1 = mz[j + 1]
                inten_j = inten[j]
                inten_j1 = inten[j + 1]
            elif mz_j > x and j > 0:
                mz_j1 = mz_j
                inten_j1 = inten[j]
                mz_j = mz[j - 1]
                inten_j = mz[j - 1]
            else:
                continue
            contrib = ((inten_j * (mz_j1 - x)) + (inten_j1 * (x - mz_j))) / (mz_j1 - mz_j)
            intensity_array[i] += contrib * weight
    return mz_array, intensity_array / sum(weights)


try:
    _has_c = True
    _average_signal = average_signal
    from ms_peak_picker._c.scan_averaging import average_signal
except ImportError:
    _has_c = False



cimport cython
cimport numpy as np
from ms_peak_picker._c.double_vector cimport DoubleVector

cdef int vgauss(double* x, size_t n, double mean, double variance, double scale, DoubleVector* out) nogil
cdef int mean_var(double* x, size_t n, double* mean, double* variance) nogil

cpdef np.ndarray[double, ndim=1, mode='c'] gaussian_smooth(
        np.ndarray[double, ndim=1, mode='c'] x,
        np.ndarray[double, ndim=1, mode='c'] y, double width=*)



cimport numpy as np
from ms_peak_picker._c.peak_set cimport PeakSet, FittedPeak


cdef class PeakIndex(object):
    cdef:
        public np.ndarray mz_array
        public np.ndarray intensity_array
        public PeakSet peaks

    cdef FittedPeak _has_peak(self, double mz, double tolerance=*)
    cdef PeakSet _between(self, double start, double stop)

    cpdef tuple all_peaks_for(self, double mz, double error_tolerance=*)

    cdef size_t get_size(self)

    cpdef PeakIndex clone(self)
    cpdef PeakIndex copy(self)


# cython: embedsignature=True

cimport cython
from cython cimport parallel
cimport numpy as np
from libc cimport math
from libc.math cimport fabs, sqrt, log, ceil, floor, log
from libc.stdlib cimport malloc, calloc, free
import numpy as np

from cpython cimport PyFloat_AsDouble
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM

np.import_array()

ctypedef double mz_type
ctypedef double intensity_type
ctypedef np.uint64_t count_type


@cython.cdivision(True)
@cython.boundscheck(False)
cdef double log2(double x) nogil:
    return log(x) / log(2)


@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint isclose(double x, double y, double rtol=1.e-5, double atol=1.e-8) nogil:
    return fabs(x-y) <= (atol + rtol * fabs(y))


@cython.cdivision(True)
@cython.boundscheck(False)
cdef size_t binsearch(double* array, double x, size_t n) nogil:
    cdef:
        size_t lo, hi, mid
        double y, err

    lo = 0
    hi = n

    while hi != lo:
        mid = (hi + lo) / 2
        y = array[mid]
        err = y - x
        if hi - lo == 1:
            return mid
        elif isclose(err, 0):
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


cdef int between_search(double *array, double lo, double hi, size_t n, ssize_t* lo_i_out, ssize_t* hi_i_out) nogil:
    cdef:
        size_t lo_i, hi_i
    lo_i = binsearch(array, lo, n)
    hi_i = binsearch(array, hi, n)

    if lo - array[lo_i] > 0.1:
        if lo_i < n - 1:
            lo_i += 1
    if ((array[hi_i] - hi) > 0.1) and (hi_i - 1) > lo_i:
        if hi_i > 0:
            hi_i -= 1
    if lo_i > hi_i:
        hi_i = lo_i
    lo_i_out[0] = lo_i
    hi_i_out[0] = hi_i
    return 0


def pybetween_search(double[:] arr, double x, double y):
    cdef:
        ssize_t lo, hi
    between_search(&arr[0], x, y, arr.size, &lo, &hi)
    return lo, hi


def pybinsearch(double[:] arr, double x):
    return binsearch(&arr[0], x, arr.size)


@cython.cdivision(True)
@cython.boundscheck(False)
cdef double percentile(intensity_type[:] N, double percent) nogil:
    cdef:
        double k, f, c, d0, d1
    k = (N.shape[0] - 1) * percent
    f = floor(k)
    c = ceil(k)
    if f == c:
        return N[<size_t>(k)]
    d0 = N[<size_t>(f)] * (c - k)
    d1 = N[<size_t>(c)] * (k - f)
    return d0 + d1


@cython.cdivision(True)
@cython.boundscheck(False)
cdef double freedman_diaconis_bin_width(intensity_type[:] x) nogil:
    cdef:
        double q75, q25, iqr

    q75 = percentile(x, 0.75)
    q25 = percentile(x, 0.25)
    iqr = q75 - q25
    return 2.0 * iqr * (x.shape[0] ** (-1.0 / 3.0))


@cython.cdivision(True)
@cython.boundscheck(False)
cdef double sturges_bin_width(intensity_type[:] x) nogil:
    cdef:
        double d
        double mn, mx
    d = log2(x.shape[0] + 1.0)
    minmax(x, &mn, &mx)
    return (mx - mn) / d


@cython.cdivision(True)
@cython.boundscheck(False)
cdef int minmax(intensity_type[:] x, intensity_type* minimum, intensity_type* maximum) nogil:
    cdef:
        size_t i, n
        intensity_type d, lo, hi

    n = x.shape[0]
    if n == 0:
        minimum[0] = 0
        maximum[0] = 0
        return 0
    lo = x[0]
    hi = x[0]

    for i in range(x.shape[0]):
        d = x[i]
        if d < lo:
            lo = d
        if d > hi:
            hi = d
    minimum[0] = lo
    maximum[0] = hi
    return 0


def pyminmax(x):
    cdef:
        intensity_type a, b
    minmax(x, &a, &b)
    return a, b


@cython.cdivision(True)
@cython.boundscheck(False)
cdef int _histogram(intensity_type[:] a, count_type* bin_count, intensity_type* bin_edges, int bins=10) nogil:
    cdef:
        size_t i, n, j
        intensity_type mn, mx, norm, x
        int hit
        intensity_type binwidth, binwidth_fd, binwidth_st

    minmax(a, &mn, &mx)

    if mn == mx:
        mn -= 0.5
        mx += 0.5

    binwidth = (mx - mn) / bins

    norm = bins / (mx - mn)

    for i in range(bins + 1):
        bin_edges[i] = i * binwidth

    n = a.shape[0]
    for i in range(n):
        x = a[i]
        hit = 0
        for j in range(1, bins + 1):
            binwidth = bin_edges[j]
            if x < binwidth:
                hit = 1
                bin_count[j - 1] += 1
                break
        if not hit:
            bin_count[j - 1] += 1
    return 0


def histogram(intensity_type[:] a, int bins=10):
    cdef:
        np.ndarray[count_type, ndim=1, mode='c'] bin_count
        np.ndarray[intensity_type, ndim=1, mode='c'] bin_edges
    bin_count = np.zeros(bins, dtype=np.uint64)
    bin_edges = np.empty(bins + 1)
    _histogram(a, &bin_count[0], &bin_edges[0], bins)
    return bin_count, bin_edges


cdef class Window(object):
    cdef:
        public mz_type[:] mz_array
        public intensity_type[:] intensity_array
        public ssize_t start_index
        public ssize_t end_index
        public mz_type center_mz
        public intensity_type mean_intensity
        public ssize_t size
        public bint is_empty
        mz_type* _mz_array
        intensity_type* _intensity_array

        int bins
        count_type* _bin_count
        intensity_type* _bin_edges

    def __init__(self, np.ndarray[mz_type, ndim=1, mode='c'] mz_array,
                 np.ndarray[intensity_type, ndim=1, mode='c'] intensity_array,
                 start_index, end_index, center_mz, bins=10, is_empty=False):
        self.mz_array = mz_array
        self.intensity_array = intensity_array
        self.start_index = start_index
        self.end_index = end_index
        self.center_mz = center_mz
        self._mz_array = &mz_array[0]
        self._intensity_array = &intensity_array[0]
        self.size = mz_array.shape[0]
        self.mean_intensity = -1
        self.bins = bins
        self.is_empty = is_empty
        self._init_arrays()
        self.deduct_intensity(0)

    def __dealloc__(self):
        self._release_arrays()

    cdef void _init_arrays(self):
        self._release_arrays()
        self._bin_count = <count_type*>calloc(sizeof(count_type), self.bins)
        self._bin_edges = <intensity_type*>malloc(sizeof(intensity_type) * (self.bins + 1))

    cdef void _release_arrays(self):
        if self._bin_edges != NULL:
            free(self._bin_edges)
            self._bin_edges = NULL
        if self._bin_count != NULL:
            free(self._bin_count)
            self._bin_count = NULL

    @cython.boundscheck(False)
    @cython.cdivision(True)
    cdef void _deduct_intensity(self, double value):
        cdef:
            size_t i, n
            double total
        total = 0
        n = self.size
        for i in range(n):
            self._intensity_array[i] -= value
            if self._intensity_array[i] < 0:
                self._intensity_array[i] = 0
            total += self._intensity_array[i]
        if n == 0:
            n = 1
        self.mean_intensity = total / n

    cpdef deduct_intensity(self, double value):
        self._deduct_intensity(value)

    @cython.boundscheck(False)
    @cython.cdivision(True)
    cpdef double truncated_mean(self, double threshold=0.95):
        cdef:
            double mask_level, total, weight
            size_t i_count, n_count
        if self.size == 0:
            return 1e-6
        _histogram(self.intensity_array, self._bin_count, self._bin_edges, self.bins)

        n_count = self.bins
        mask_level = 0
        for i_count in range(n_count):
            if self._bin_count[i_count] > mask_level:
                mask_level = self._bin_count[i_count]
        mask_level *= (1 - threshold)
        total = 0
        weight = 0
        for i_count in range(n_count):
            if mask_level < self._bin_count[i_count]:
                total += self._bin_edges[i_count + 1] * self._bin_count[i_count]
                weight += self._bin_count[i_count]
        return total / weight

    def __repr__(self):
        return "Window(start_index=%d, end_index=%d, mean_intensity=%f, center_mz=%f)" % (
            self.start_index, self.end_index, self.mean_intensity, self.center_mz)


cdef class NoiseRegion(object):
    cdef:
        public list windows
        public double width
        public size_t start_index
        public size_t end_index
        public size_t size


    def __init__(self, windows, width=10):
        self.windows = windows
        self.width = width
        self.start_index = windows[0].start_index
        self.end_index = windows[-1].end_index
        self.size = PyList_GET_SIZE(windows)

    def __repr__(self):
        return "NoiseRegion(start_index=%d, end_index=%d)" % (
            self.start_index, self.end_index)

    cpdef Window noise_window(self):
        cdef:
            size_t i, n
            double minimum
            Window minimum_window, window

        i = 0
        n = self.size
        if n == 0:
            return None
        minimum_window = self.getitem(0)
        minimum = minimum_window.mean_intensity
        for i in range(1, n):
            window = self.getitem(i)
            if window.mean_intensity < minimum:
                minimum_window = window
                minimum = minimum_window.mean_intensity
        return minimum_window

    def arrays(self):
        cdef:
            size_t i, n
            Window w
        mz_array = []
        intensity_array = []
        n = self.size
        for i in range(n):
            w = self.getitem(i)
            if w.is_empty:
                continue
            mz_array.append(w.mz_array)
            intensity_array.append(w.intensity_array)
        if len(mz_array) > 0:
            mz_array = np.concatenate(mz_array)
            intensity_array = np.concatenate(intensity_array)
        else:
            mz_array = np.array([])
            intensity_array = np.array([])
        return mz_array, intensity_array

    @cython.final
    cdef Window getitem(self, size_t i):
        return <Window>PyList_GET_ITEM(self.windows, i)

    cpdef double denoise(self, double scale=5, int maxiter=10):
        cdef:
            double noise_mean, last_mean
            size_t i, n, niter

        if scale == 0:
            return 0
        noise_mean = self.noise_window().truncated_mean() * scale
        n = self.size
        for i in range(n):
            window = self.getitem(i)
            window.deduct_intensity(noise_mean)
        last_mean = noise_mean
        noise_mean = self.noise_window().truncated_mean()
        niter = 1
        while abs(last_mean - noise_mean) > 1e-3 and niter < maxiter:
            niter += 1
            noise_mean = self.noise_window().truncated_mean() * scale
            for i in range(n):
                window = self.getitem(i)
                window._deduct_intensity(noise_mean)
            last_mean = noise_mean

        return last_mean - noise_mean


cpdef list windowed_spectrum(np.ndarray[mz_type, ndim=1, mode='c'] mz_array,
                             np.ndarray[intensity_type, ndim=1, mode='c'] intensity_array,
                             double window_size=1.):
    cdef:
        mz_type mz_min, mz_max, step_size, center_mz
        mz_type lo_mz, hi_mz
        ssize_t center_i, niter, lo_i, hi_i, i, n
        ssize_t last_lo_i, last_hi_i
        list windows 
        Window win
        mz_type* _mz_array

    n = mz_array.shape[0]
    _mz_array = &mz_array[0]
    if n < 2:
        return []
    mz_min = _mz_array[0]
    mz_max = _mz_array[n - 1]


    step_size = window_size / 2.
    center_mz = mz_min + step_size
    center_i = binsearch(_mz_array, center_mz, n)

    windows = []

    niter = 0
    last_lo_i = 0
    last_hi_i = 0
    while center_mz < mz_max:
        lo_mz = center_mz - step_size
        hi_mz = center_mz + step_size
        # find the indices that bound to lo_mz and hi_mz, but which
        # do not exceed 
        between_search(_mz_array, lo_mz, hi_mz, n, &lo_i, &hi_i)
        # if the found boundaries' average m/z is contained in the
        # theoretical boundary m/z range, this window is populated
        if lo_mz <= (_mz_array[lo_i] + _mz_array[hi_i]) / 2 <= hi_mz:        
            win = Window(mz_array[lo_i:hi_i + 1],
                         intensity_array[lo_i:hi_i + 1],
                         lo_i, hi_i, center_mz, is_empty=False)
        # otherwise there was no signal in this range and it is empty
        else:
            win = Window(np.array([center_mz], dtype=np.float64),
                         np.array([0.0], dtype=np.float64), lo_i, hi_i,
                         center_mz, is_empty=True)
        last_lo_i = lo_i
        last_hi_i = hi_i
        windows.append(win)

        center_mz = center_mz + window_size
        center_i = binsearch(_mz_array, center_mz, n)

        niter += 1

    return windows


cpdef list filter_windows(list windows):
    cdef:
        size_t i, n
        Window window
        list filtered
    filtered = []
    n = PyList_GET_SIZE(windows)

    for i in range(n):
        window = <Window>PyList_GET_ITEM(windows, i)
        if not window.is_empty:
            filtered.append(window)
    return filtered


cpdef list group_windows(list windows, int width=10):
    cdef:
        int step
        list regions, selected_windows
        size_t i, n

    step = int(width / 2)
    regions = []
    i = step
    n = len(windows)
    while i < n:
        lo = i - step
        hi = i + step

        selected_windows = (windows[lo:hi])
        # if PyList_GET_SIZE(selected_windows) > 0:
        #     reg = NoiseRegion(selected_windows)
        #     regions.append(reg)
        reg = NoiseRegion(selected_windows)
        regions.append(reg)
        i += step * 2
    return regions


cdef class FTICRScan(object):
    cdef:
        public np.ndarray mz_array
        public np.ndarray intensity_array
        public list windows
        public list regions

    def __init__(self, mz_array, intensity_array):
        self.mz_array = mz_array.astype(float)
        self.intensity_array = intensity_array.astype(float)

        self.windows = None
        self.regions = None

    def __iter__(self):
        yield self.mz_array
        yield self.intensity_array

    def build_windows(self, window_size=1.):
        self.windows = windowed_spectrum(
            self.mz_array, self.intensity_array, window_size=window_size)

    def build_regions(self, region_width=10):
        self.regions = group_windows(self.windows, region_width)

    def denoise(self, window_size=1., region_width=10, scale=5):
        cdef:
            size_t i, n
            NoiseRegion region
            list mz, intensity
            object mz_, intensity_
        self.build_windows(window_size)
        self.build_regions(region_width)
        n = len(self.regions)
        for i in range(n):
            region = self.regions[i]
            # Modify the windowed arrays in-place
            region.denoise(scale)
        return self



# cython: embedsignature=True

cimport cython
from cython cimport parallel
cimport numpy as np
from libc cimport math
import numpy as np

from ms_peak_picker._c.search cimport get_nearest
from ms_peak_picker._c.double_vector cimport (
    DoubleVector, make_double_vector_with_size,
    make_double_vector, double_vector_resize,
    double_vector_append, free_double_vector,
    print_double_vector, double_vector_to_list,
    list_to_double_vector)

from ms_peak_picker._c.peak_set cimport FittedPeak

from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM

from ms_peak_picker._c.peak_set cimport FittedPeak
from ms_peak_picker._c.search cimport get_nearest_binary

np.import_array()

cdef DTYPE_t minimum_signal_to_noise = 4.0


@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint isclose(DTYPE_t x, DTYPE_t y, DTYPE_t rtol=1.e-5, DTYPE_t atol=1.e-8) nogil:
    return math.fabs(x-y) <= (atol + rtol * math.fabs(y))


@cython.nonecheck(False)
@cython.cdivision(True)
cdef bint aboutzero(DTYPE_t x) nogil:
    return isclose(x, 0)


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef DTYPE_t find_signal_to_noise(double target_val, np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array, size_t index):
    cdef:
        DTYPE_t min_intensity_left, min_intensity_right
        size_t size, i
        DTYPE_t* pintensity_array
        int path_id

    with nogil:
        min_intensity_left = 0
        min_intensity_right = 0
        size = intensity_array.shape[0] - 1
        if aboutzero(target_val):
            return 0
        if index <= 0 or index >= size:
            return 0

        pintensity_array = &intensity_array[0]

        # locate the next valley to the left
        for i in range(index, 0, -1):
            if pintensity_array[i + 1] >= pintensity_array[i] and pintensity_array[i - 1] > pintensity_array[i]:
                min_intensity_left = pintensity_array[i]
                break
        else:
            min_intensity_left = pintensity_array[0]
        # locate the next valley to the right
        for i in range(index, size):
            if pintensity_array[i + 1] >= pintensity_array[i] and pintensity_array[i - 1] > pintensity_array[i]:
                min_intensity_right = pintensity_array[i]
                break
        else:
            min_intensity_right = pintensity_array[size]

        if aboutzero(min_intensity_left):
            if aboutzero(min_intensity_right):
                return 100
            else:
                return target_val / min_intensity_right

        if min_intensity_right < min_intensity_left and not aboutzero(min_intensity_right):
            return target_val / min_intensity_right
        return target_val / min_intensity_left


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef DTYPE_t curve_reg(np.ndarray[DTYPE_t, ndim=1, mode='c'] x, np.ndarray[DTYPE_t, ndim=1, mode='c'] y, size_t n,
                        np.ndarray[DTYPE_t, ndim=1, mode='c'] terms, size_t nterms):
    """
    Fit a least squares polynomial regression

    Parameters
    ----------
    x : array
    y : array
    n : int
    terms : array
        Mutated to pass back coefficients
        of fit.
    nterms : int
        Number of terms

    Returns
    -------
    float
    """
    cdef:
        np.ndarray[DTYPE_t, ndim=1] weights
        np.ndarray[DTYPE_t, ndim=2] At, Z, At_T, At_At_T, I_At_At_T, At_Ai_At
        np.ndarray[DTYPE_t, ndim=2] B, out
        DTYPE_t mse, yfit, xpow
        size_t i, j

    weights = np.ones(n)

    # Weighted powers of x transposed
    # the polynomial regression's design matrix
    At = np.zeros((nterms + 1, n))
    for i in range(n):
        # set the intercept
        At[0, i] = weights[i]
        for j in range(1, nterms + 1):
            # the successive powers of x[i]
            At[j, i] = At[j - 1, i] * x[i]

    Z = np.zeros((n, 1))
    for i in range(n):
        Z[i, 0] = weights[i] * y[i]

    At_T = At.T
    At_At_T = At.dot(At_T)
    I_At_At_T = np.linalg.inv(At_At_T)
    At_Ai_At = I_At_At_T.dot(At)
    B = At_Ai_At.dot(Z)
    mse = 0
    out = np.zeros((2, n))
    for i in range(n):
        terms[0] = B[0, 0]
        yfit = B[0, 0]
        xpow = x[i]
        for j in range(1, nterms):
            terms[j] = B[j, 0]
            yfit += B[j, 0] * xpow
            xpow = xpow * x[i]
        out[0, i] = yfit
        out[1, i] = y[i] - yfit
        mse += y[i] - yfit
    return mse



@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cdef DTYPE_t curve_reg_dv(DoubleVector* x, DoubleVector* y, size_t n,
                        np.ndarray[DTYPE_t, ndim=1, mode='c'] terms, size_t nterms):
    cdef:
        np.ndarray[DTYPE_t, ndim=1] weights
        np.ndarray[DTYPE_t, ndim=2] At, Z, At_Ai_At
        np.ndarray[DTYPE_t, ndim=2] At_T, At_At_T, I_At_At_T
        np.ndarray[DTYPE_t, ndim=2] B, out
        DTYPE_t mse, yfit, xpow
        size_t i, j

    weights = np.ones(n)

    # Weighted powers of x transposed
    # Like Vandermonte Matrix?
    At = np.zeros((nterms + 1, n))
    for i in range(n):
        At[0, i] = weights[i]
        for j in range(1, nterms + 1):
            At[j, i] = At[j - 1, i] * x.v[i]

    Z = np.zeros((n, 1))
    for i in range(n):
        Z[i, 0] = weights[i] * y.v[i]

    At_T = At.T
    At_At_T = At.dot(At_T)
    try:
        I_At_At_T = np.linalg.inv(At_At_T)
    except Exception:
        for i in range(n):
            terms[i] = 0
        return -1
    At_Ai_At = I_At_At_T.dot(At)
    B = At_Ai_At.dot(Z)
    mse = 0
    out = np.zeros((2, n))
    for i in range(n):
        terms[0] = B[0, 0]
        yfit = B[0, 0]
        xpow = x.v[i]
        for j in range(1, nterms):
            terms[j] = B[j, 0]
            yfit += B[j, 0] * xpow
            xpow = xpow * x.v[i]
        out[0, i] = yfit
        out[1, i] = y.v[i] - yfit
        mse += y.v[i] - yfit
    return mse


@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef DTYPE_t find_right_width(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array, np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array,
                               size_t data_index, DTYPE_t signal_to_noise=0.):
    cdef:
        int points
        DTYPE_t peak, peak_half, mass, X1, X2, Y1, Y2, mse, last_Y1
        DTYPE_t lower, current_mass
        size_t size, index, j, k
        np.ndarray[DTYPE_t, ndim=1, mode='c'] coef
        DoubleVector *vect_mzs
        DoubleVector *vect_intensity
    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if peak == 0.0:
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.

    last_Y1 = peak
    lower = mz_array[size]
    for index in range(data_index, size):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if((Y1 < peak_half) or (abs(mass - current_mass) > 1.5) or (Y1 > last_Y1) or (
                (index > size - 1 or intensity_array[index + 1] > Y1) and (
                index > size - 2 or intensity_array[index + 2] > Y1) and signal_to_noise < minimum_signal_to_noise)):
            Y2 = intensity_array[index - 1]
            X1 = mz_array[index]
            X2 = mz_array[index - 1]

            if((Y2 - Y1 != 0) and (Y1 < peak_half)):
                lower = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                lower = X1
                points = index - data_index + 1

                if points >= 3:
                    vect_mzs = make_double_vector()
                    vect_intensity = make_double_vector()

                    for k in range(points - 1, -1, -1):
                        double_vector_append(vect_mzs, mz_array[index - k])
                        double_vector_append(vect_intensity, intensity_array[index - k])
                    j = 0
                    while (j < points) and (vect_intensity.v[0] == vect_intensity.v[j]):
                        j += 1

                    if j == points:
                        return 0.0

                    # coef will contain the result
                    mse = curve_reg_dv(vect_intensity, vect_mzs, points, coef, 1)
                    free_double_vector(vect_intensity)
                    free_double_vector(vect_mzs)
                    lower = coef[1] * peak_half + coef[0]
            break
        last_Y1 = Y1
    return abs(lower - mass)



@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef DTYPE_t find_left_width(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array, np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array,
                              size_t data_index, DTYPE_t signal_to_noise=0.):
    cdef:
        int points
        DTYPE_t peak, peak_half, mass, X1, X2, Y1, Y2, mse, last_Y1
        DTYPE_t upper, current_mass
        size_t size, index, j, k
        np.ndarray[DTYPE_t, ndim=1, mode='c'] coef
        DoubleVector *vect_mzs
        DoubleVector *vect_intensity

    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if peak == 0.0:
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.
    last_Y1 = peak
    upper = mz_array[0]
    for index in range(data_index, -1, -1):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if ((Y1 < peak_half) or (abs(mass - current_mass) > 1.5) or (Y1 > last_Y1) or (
                (index < 1 or intensity_array[index - 1] > Y1) and (
                    index < 2 or intensity_array[index - 2] > Y1) and (signal_to_noise < minimum_signal_to_noise))):
            Y2 = intensity_array[index + 1]
            X1 = mz_array[index]
            X2 = mz_array[index + 1]

            if ((Y2 - Y1 != 0) and (Y1 < peak_half)):
                upper = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                upper = X1
                points = data_index - index + 1
                if points >= 3:
                    vect_mzs = make_double_vector()
                    vect_intensity = make_double_vector()

                    for j in range(points - 1, -1, -1):
                        double_vector_append(vect_mzs, mz_array[data_index - j])
                        double_vector_append(vect_intensity, intensity_array[data_index - j])

                    j = 0
                    while j < points and isclose(vect_intensity.v[0], vect_intensity.v[j]):
                        j += 1

                    if j == points:
                        return 0.

                    # coef will contain the results
                    mse = curve_reg_dv(vect_intensity, vect_mzs, points, coef, 1)
                    free_double_vector(vect_intensity)
                    free_double_vector(vect_mzs)
                    upper = coef[1] * peak_half + coef[0]
            break
        last_Y1 = Y1
    return abs(mass - upper)



@cython.nonecheck(False)
@cython.cdivision(True)
@cython.boundscheck(False)
cpdef DTYPE_t find_full_width_at_half_max(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array, np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array,
                                          size_t data_index, double signal_to_noise=0.):
    cdef:
        int points
        DTYPE_t peak, peak_half, mass, X1, X2, Y1, Y2, mse
        DTYPE_t upper, lower, current_mass
        size_t size, index, j, k
        np.ndarray[DTYPE_t, ndim=1, mode='c'] coef
        DoubleVector* vect_mzs
        DoubleVector* vect_intensity

    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if aboutzero(peak):
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.

    upper = mz_array[0]
    for index in range(data_index, -1, -1):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if ((Y1 < peak_half) or (abs(mass - current_mass) > 1.5) or (
                (index < 1 or intensity_array[index - 1] > Y1) and (
                 index < 2 or intensity_array[index - 2] > Y1) and (signal_to_noise < minimum_signal_to_noise))):
            Y2 = intensity_array[index + 1]
            X1 = mz_array[index]
            X2 = mz_array[index + 1]

            if (not aboutzero(Y2 - Y1) and (Y1 < peak_half)):
                upper = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                upper = X1
                points = data_index - index + 1
                if points >= 3:

                    vect_mzs = make_double_vector()
                    vect_intensity = make_double_vector()

                    for j in range(points - 1, -1, -1):
                        double_vector_append(vect_mzs, mz_array[data_index - j])
                        double_vector_append(vect_intensity, intensity_array[data_index - j])

                    j = 0
                    while j < points and (vect_intensity.v[0] == vect_intensity.v[j]):
                        j += 1

                    if j == points:
                        return 0.
                    mse = curve_reg_dv(vect_intensity, vect_mzs, points, coef, 1)
                    upper = coef[1] * peak_half + coef[0]
                    free_double_vector(vect_mzs)
                    free_double_vector(vect_intensity)
            break
    lower = mz_array[size]
    for index in range(data_index, size):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if((Y1 < peak_half) or (abs(mass - current_mass) > 1.5) or ((index > size - 1 or intensity_array[index + 1] > Y1) and (
                    index > size - 2 or intensity_array[index + 2] > Y1) and signal_to_noise < 4.0)):
            Y2 = intensity_array[index - 1]
            X1 = mz_array[index]
            X2 = mz_array[index - 1]

            if(not aboutzero(Y2 - Y1) and (Y1 < peak_half)):
                lower = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                lower = X1
                points = index - data_index + 1
                if points >= 3:
                    vect_mzs = make_double_vector()
                    vect_intensity = make_double_vector()

                    for k in range(points - 1, -1, -1):
                        double_vector_append(vect_mzs, mz_array[index - k])
                        double_vector_append(vect_intensity, intensity_array[index - k])
                    j = 0
                    while (j < points) and (vect_intensity.v[0] == vect_intensity.v[j]):
                        j += 1

                    if j == points:
                        return 0.0
                    mse = curve_reg_dv(vect_intensity, vect_mzs, points, coef, 1)
                    free_double_vector(vect_intensity)
                    free_double_vector(vect_mzs)
                    lower = coef[1] * peak_half + coef[0]
            break

    if aboutzero(upper):
        return 2 * abs(mass - lower)
    if aboutzero(lower):
        return 2 * abs(mass - upper)
    return abs(upper - lower)


@cython.boundscheck(False)
@cython.cdivision(True)
cpdef DTYPE_t quadratic_fit(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array,
                            np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array,
                            ssize_t index):
    cdef:
        DTYPE_t x1, x2, x3
        DTYPE_t y1, y2, y3
        DTYPE_t d, mz_fit

    if index < 1:
        return mz_array[0]
    elif index > mz_array.shape[0] - 1:
        return mz_array[-1]
    x1 = mz_array[index - 1]
    x2 = mz_array[index]
    x3 = mz_array[index + 1]
    y1 = intensity_array[index - 1]
    y2 = intensity_array[index]
    y3 = intensity_array[index + 1]

    d = (y2 - y1) * (x3 - x2) - (y3 - y2) * (x2 - x1)
    if d == 0:  # If the interpolated intensity is 0, the peak fitting is no better than the peak
        return x2
    mz_fit = ((x1 + x2) - ((y2 - y1) * (x3 - x2) * (x1 - x3)) / d) / 2.0
    return mz_fit


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef double lorentzian_least_squares(DTYPE_t[:] mz_array,
                                     DTYPE_t[:] intensity_array,
                                     double amplitude, double full_width_at_half_max,
                                     double vo, size_t lstart, size_t lstop) nogil:

    cdef:
        double root_mean_squared_error, u, Y2
        long Y1
        size_t index


    root_mean_squared_error = 0

    for index in range(lstart, lstop + 1):
        u = 2 / float(full_width_at_half_max) * (mz_array[index] - vo)
        Y1 = int(amplitude / float(1 + u * u))
        Y2 = intensity_array[index]

        root_mean_squared_error += (Y1 - Y2) * (Y1 - Y2)

    return root_mean_squared_error


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef double lorentzian_fit(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array, np.ndarray[DTYPE_t, ndim=1, mode='c'] intensity_array,
                            size_t index, double full_width_at_half_max):
    cdef:
        double amplitude
        DTYPE_t vo, step, current_error, last_error
        size_t lstart, lstop, i
        DTYPE_t[:] view_mz_array
        DTYPE_t[:] view_intensity_array

    amplitude = intensity_array[index]
    vo = mz_array[index]
    step = math.fabs((vo - mz_array[index + 1]) / 100.0)

    if index < 1:
        return mz_array[index]
    elif index >= mz_array.shape[0] - 1:
        return mz_array[-1]

    lstart = get_nearest(mz_array, vo + full_width_at_half_max, index) + 1
    lstop = get_nearest(mz_array, vo - full_width_at_half_max, index) - 1

    view_mz_array = mz_array
    view_intensity_array = intensity_array

    with nogil:
        current_error = lorentzian_least_squares(
            view_mz_array, view_intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
        for i in range(50):
            last_error = current_error
            vo = vo + step
            current_error = lorentzian_least_squares(
                view_mz_array, view_intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
            if (current_error > last_error):
                break

        vo = vo - step
        current_error = lorentzian_least_squares(
            view_mz_array, view_intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
        for i in range(50):
            last_error = current_error
            vo = vo - step
            current_error = lorentzian_least_squares(
                view_mz_array, view_intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
            if (current_error > last_error):
                break

        vo += step
        return vo


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef double peak_area(np.ndarray[DTYPE_t, ndim=1, mode='c'] mz_array, np.ndarray[DTYPE_t, ndim=1, mode='c']  intensity_array,
                       size_t start, size_t stop):
    cdef:
        double area
        size_t i
        DTYPE_t x1, y1, x2, y2

    area = 0.
    for i in range(start + 1, stop):
        x1 = mz_array[i - 1]
        y1 = intensity_array[i - 1]
        x2 = mz_array[i]
        y2 = intensity_array[i]
        area += (y1 * (x2 - x1)) + ((y2 - y1) * (x2 - x1) / 2.)

    return area


@cython.cdivision(True)
cpdef double gaussian_predict(FittedPeak peak, double mz):
    cdef:
        double x, center, amplitude, fwhm, spread, y
    x = mz
    center = peak.mz
    amplitude = peak.intensity
    fwhm = peak.full_width_at_half_max
    spread = fwhm / 2.35482
    y = amplitude * math.exp(-((x - center) ** 2) / (2 * spread ** 2))
    return y


cpdef object gaussian_shape(FittedPeak peak):
    cdef:
        double center, amplitude, fwhm, spread
        np.ndarray[double, ndim=1] x, y
    center = peak.mz
    amplitude = peak.intensity
    fwhm = peak.full_width_at_half_max
    spread = fwhm / 2.35482
    x = np.arange(center - fwhm - 0.02, center + fwhm + 0.02, 0.0001)
    y = amplitude * np.exp(-((x - center) ** 2) / (2 * spread ** 2))
    return x, y


cpdef double gaussian_error(FittedPeak peak, double mz, double intensity):
    y = gaussian_predict(peak, mz)
    return intensity - y


cpdef double gaussian_volume(FittedPeak peak):
    cdef:
        np.ndarray[double, ndim=1] x, y
    x, y = gaussian_shape(peak)
    return np.trapz(y, x, dx=0.0001)


cdef class PeakShapeModel(object):

    def __init__(self, peak):
        self.peak = peak
        self.center = peak.mz

    cpdef double predict(self, double mz):
        raise NotImplementedError()

    cpdef object shape(self):
        raise NotImplementedError()

    cpdef double volume(self):
        raise NotImplementedError()

    cpdef double error(self, double mz, double intensity):
        raise NotImplementedError()

    def __repr__(self):
        return "{self.__class__.__name__}({self.peak})".format(self=self)


cdef class GaussianModel(PeakShapeModel):
    cpdef object shape(self):
        return gaussian_shape(self.peak)

    cpdef double predict(self, double mz):
        return gaussian_predict(self.peak, mz)

    cpdef double volume(self):
        return gaussian_volume(self.peak)

    cpdef double error(self, double mz, double intensity):
        return gaussian_error(self.peak, mz, intensity)


cdef class PeakSetReprofiler(object):

    def __init__(self, models, dx=0.01):
        self.models = sorted(models, key=lambda x: x.center)
        self.dx = dx
        self.gridx = None
        self.gridy = None
        self._build_grid()

    cdef void _build_grid(self):
        lo = self.models[0].center
        hi = self.models[-1].center
        self.gridx = np.arange(max(lo - 3, 0), hi + 3, self.dx, dtype=np.float64)
        self.gridy = np.zeros_like(self.gridx, dtype=np.float64)

    @cython.boundscheck(False)
    cpdef _reprofile(self):
        cdef:
            ssize_t i, j, nmodels, offset
            double x, y, pred
            PeakShapeModel model
            np.ndarray[double, ndim=1, mode='c'] gridx, gridy
            list models
        gridx = self.gridx
        gridy = self.gridy
        models = self.models
        nmodels = PyList_GET_SIZE(models)
        i = 0
        for i in range(gridx.shape[0]):
            x = gridx[i]
            y = 0
            offset = self._find_starting_model(x)
            j = offset - 1
            while j > 0:
                model = <PeakShapeModel>PyList_GET_ITEM(models, j)
                if (x - model.center) > 3:
                    break
                pred = model.predict(x)
                y += pred
                j -= 1
            j = offset
            while j < nmodels:
                model = <PeakShapeModel>PyList_GET_ITEM(models, j)
                if (model.center - x) > 3:
                    break
                pred = model.predict(x)
                y += pred
                j += 1

            gridy[i] = y
        gridx = self.gridx
        gridy = self.gridy

    cpdef size_t _find_starting_model(self, double x):
        cdef:
            size_t lo, hi, mid
            list models
            double center, err
            PeakShapeModel model
        lo = 0
        models = self.models
        hi = PyList_GET_SIZE(models)
        while (lo != hi):
            mid = (hi + lo) // 2
            model = <PeakShapeModel>PyList_GET_ITEM(models, mid)
            center = model.center
            err = center - x
            if abs(err) < 0.0001:
                return mid
            elif (hi - lo) == 1:
                return mid
            elif err > 0:
                hi = mid
            else:
                lo = mid
        return 0

    def reprofile(self):
        self._reprofile()
        return self.gridx, self.gridy



import numpy as np

from .utils import range
from .search import get_nearest


minimum_signal_to_noise = 4.


def find_signal_to_noise(target_val, intensity_array, index):
    min_intensity_left = 0
    min_intensity_right = 0
    size = len(intensity_array) - 1
    if target_val == 0:
        return 0
    if index <= 0 or index >= size:
        return 0

    for i in range(index, 0, -1):
        if intensity_array[i + 1] >= intensity_array[i] and intensity_array[i - 1] > intensity_array[i]:
            min_intensity_left = intensity_array[i]
            break
    else:
        min_intensity_left = intensity_array[0]

    for i in range(index, size):
        if intensity_array[i + 1] >= intensity_array[i] and intensity_array[i - 1] > intensity_array[i]:
            min_intensity_right = intensity_array[i]
            break
    else:
        min_intensity_right = intensity_array[size]

    if min_intensity_left == 0:
        if min_intensity_right == 0:
            return 100
        else:
            return target_val / min_intensity_right

    if min_intensity_right < min_intensity_left and min_intensity_right != 0:
        return target_val / min_intensity_right
    return target_val / min_intensity_left


def quadratic_fit(mz_array, intensity_array, index):
    if index < 1:
        return mz_array[0]
    elif index > len(mz_array) - 1:
        return mz_array[-1]
    x1, x2, x3 = mz_array[(index - 1):(index + 2)]
    y1, y2, y3 = intensity_array[(index - 1):(index + 2)]

    d = (y2 - y1) * (x3 - x2) - (y3 - y2) * (x2 - x1)
    if d == 0:  # If the interpolated intensity is 0, the peak fitting is no better than the peak
        return x2
    mz_fit = ((x1 + x2) - ((y2 - y1) * (x3 - x2) * (x1 - x3)) / d) / 2.0
    return mz_fit


def curve_reg(x, y, n, terms, nterms):
    """
    Fit a least squares polynomial regression

    Parameters
    ----------
    x : array
    y : array
    n : int
    terms : array
        Mutated to pass back coefficients
        of fit.
    nterms : int
        Number of terms

    Returns
    -------
    float
    """
    weights = np.ones(n)

    # Weighted powers of x transposed
    At = np.zeros((nterms + 1, n))
    for i in range(n):
        # set the intercept term
        At[0, i] = weights[i]
        for j in range(1, nterms + 1):
            At[j, i] = At[j - 1, i] * x[i]

    Z = np.zeros((n, 1))
    for i in range(n):
        Z[i, 0] = weights[i] * y[i]

    # ((AtA)^-1)At
    At_T = At.T
    At_At_T = At.dot(At_T)
    I_At_At_T = np.linalg.inv(At_At_T)
    At_Ai_At = I_At_At_T.dot(At)

    B = At_Ai_At.dot(Z)

    mse = 0
    out = np.zeros((2, n))
    for i in range(n):
        terms[0] = B[0, 0]
        yfit = B[0, 0]
        xpow = x[i]
        for j in range(1, nterms):
            terms[j] = B[j, 0]
            yfit += B[j, 0] * xpow
            xpow = xpow * x[i]
        out[0, i] = yfit
        out[1, i] = y[i] - yfit
        mse += y[i] - yfit

    return mse


def find_right_width(mz_array, intensity_array, data_index, signal_to_noise=0.):
    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if peak == 0.0:
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.

    last_Y1 = peak

    lower = mz_array[size]
    for index in range(data_index, size):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if((Y1 < peak_half) or (np.fabs(mass - current_mass) > 1.5) or (Y1 > last_Y1) or (
                (index > size - 1 or intensity_array[index + 1] > Y1) and (
                index > size - 2 or intensity_array[index + 2] > Y1) and signal_to_noise < minimum_signal_to_noise)):
            Y2 = intensity_array[index - 1]
            X1 = mz_array[index]
            X2 = mz_array[index - 1]

            if((Y2 - Y1 != 0) and (Y1 < peak_half)):
                lower = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                lower = X1
                points = index - data_index + 1

                if points >= 3:
                    vect_mzs = []
                    vect_intensity = []

                    for k in range(points - 1, -1, -1):
                        vect_mzs.append(mz_array[index - k])
                        vect_intensity.append(intensity_array[index - k])
                    j = 0
                    while (j < points) and (vect_intensity[0] == vect_intensity[j]):
                        j += 1

                    if j == points:
                        return 0.0

                    # coef will contain the result
                    curve_reg(vect_intensity, vect_mzs, points, coef, 1)
                    lower = coef[1] * peak_half + coef[0]
            break
        last_Y1 = Y1
    return abs(lower - mass)


def find_left_width(mz_array, intensity_array, data_index, signal_to_noise=0.):
    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if peak == 0.0:
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.

    last_Y1 = peak
    upper = mz_array[0]
    for index in range(data_index, -1, -1):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]

        if ((Y1 < peak_half) or (np.fabs(mass - current_mass) > 1.5) or (Y1 > last_Y1) or (
                (index < 1 or intensity_array[index - 1] > Y1) and (
                    index < 2 or intensity_array[index - 2] > Y1) and (signal_to_noise < minimum_signal_to_noise))):
            Y2 = intensity_array[index + 1]
            X1 = mz_array[index]
            X2 = mz_array[index + 1]

            if ((Y2 - Y1 != 0) and (Y1 < peak_half)):
                upper = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                upper = X1
                points = data_index - index + 1
                if points >= 3:
                    vect_mzs = []
                    vect_intensity = []

                    for j in range(points - 1, -1, -1):
                        vect_mzs.append(mz_array[data_index - j])
                        vect_intensity.append(intensity_array[data_index - j])

                    j = 0
                    while j < points and (vect_intensity[0] == vect_intensity[j]):
                        j += 1

                    if j == points:
                        return 0.

                    # coef will contain the results
                    curve_reg(vect_intensity, vect_mzs, points, coef, 1)
                    upper = coef[1] * peak_half + coef[0]
            break
        last_Y1 = Y1
    return abs(mass - upper)


def find_full_width_at_half_max(mz_array, intensity_array, data_index, signal_to_noise=0.):
    points = 0
    peak = intensity_array[data_index]
    peak_half = peak / 2.
    mass = mz_array[data_index]

    coef = np.zeros(2)

    if peak == 0.0:
        return 0.

    size = len(mz_array) - 1
    if data_index <= 0 or data_index >= size:
        return 0.

    upper = mz_array[0]
    for index in range(data_index, -1, -1):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if ((Y1 < peak_half) or (np.fabs(mass - current_mass) > 1.5) or (
                (index < 1 or intensity_array[index - 1] > Y1) and (
                    index < 2 or intensity_array[index - 2] > Y1) and (signal_to_noise < minimum_signal_to_noise))):
            Y2 = intensity_array[index + 1]
            X1 = mz_array[index]
            X2 = mz_array[index + 1]

            if ((Y2 - Y1 != 0) and (Y1 < peak_half)):
                upper = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                upper = X1
                points = data_index - index + 1
                if points >= 3:
                    vect_mzs = []
                    vect_intensity = []

                    for j in range(points - 1, -1, -1):
                        vect_mzs.append(mz_array[data_index - j])
                        vect_intensity.append(intensity_array[data_index - j])

                    j = 0
                    while j < points and (vect_intensity[0] == vect_intensity[j]):
                        j += 1

                    if j == points:
                        return 0.

                    # coef will contain the results
                    curve_reg(vect_intensity, vect_mzs, points, coef, 1)
                    upper = coef[1] * peak_half + coef[0]
            break

    lower = mz_array[size]
    for index in range(data_index, size):
        current_mass = mz_array[index]
        Y1 = intensity_array[index]
        if((Y1 < peak_half) or (np.fabs(mass - current_mass) > 1.5) or (
                (index > size - 1 or intensity_array[index + 1] > Y1) and (
                index > size - 2 or intensity_array[index + 2] > Y1) and signal_to_noise < 4.0)):
            Y2 = intensity_array[index - 1]
            X1 = mz_array[index]
            X2 = mz_array[index - 1]

            if((Y2 - Y1 != 0) and (Y1 < peak_half)):
                lower = X1 - (X1 - X2) * (peak_half - Y1) / (Y2 - Y1)
            else:
                lower = X1
                points = index - data_index + 1

                if points >= 3:
                    vect_mzs = []
                    vect_intensity = []

                    for k in range(points - 1, -1, -1):
                        vect_mzs.append(mz_array[index - k])
                        vect_intensity.append(intensity_array[index - k])
                    j = 0
                    while (j < points) and (vect_intensity[0] == vect_intensity[j]):
                        j += 1

                    if j == points:
                        return 0.0

                    # coef will contain the result
                    curve_reg(vect_intensity, vect_mzs, points, coef, 1)
                    lower = coef[1] * peak_half + coef[0]
            break

    if upper == 0.0:
        return 2 * np.fabs(mass - lower)
    if lower == 0.0:
        return 2 * np.fabs(mass - upper)
    return np.fabs(upper - lower)


def lorentzian_least_squares(mz_array, intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop):
    root_mean_squared_error = 0

    for index in range(lstart, lstop + 1):
        u = 2 / float(full_width_at_half_max) * (mz_array[index] - vo)
        Y1 = int(amplitude / float(1 + u * u))
        Y2 = intensity_array[index]

        root_mean_squared_error += (Y1 - Y2) * (Y1 - Y2)

    return root_mean_squared_error


def lorentzian_fit(mz_array, intensity_array, index, full_width_at_half_max):
    amplitude = intensity_array[index]
    vo = mz_array[index]
    step = np.fabs((vo - mz_array[index + 1]) / 100.0)

    if index < 1:
        return mz_array[index]
    elif index >= len(mz_array) - 1:
        return mz_array[-1]

    lstart = get_nearest(mz_array, vo + full_width_at_half_max, index) + 1
    lstop = get_nearest(mz_array, vo - full_width_at_half_max, index) - 1
    last_error = 0.
    current_error = lorentzian_least_squares(
        mz_array, intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
    for i in range(50):
        last_error = current_error
        vo = vo + step
        current_error = lorentzian_least_squares(
            mz_array, intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
        if (current_error > last_error):
            break

    vo = vo - step
    current_error = lorentzian_least_squares(
        mz_array, intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
    for i in range(50):
        last_error = current_error
        vo = vo - step
        current_error = lorentzian_least_squares(
            mz_array, intensity_array, amplitude, full_width_at_half_max, vo, lstart, lstop)
        if (current_error > last_error):
            break

    vo += step
    return vo


def gaussian_shape(peak):
    center = peak.mz
    amplitude = peak.intensity
    fwhm = peak.full_width_at_half_max
    spread = fwhm / 2.35482
    x = np.arange(center - fwhm - 0.02, center + fwhm + 0.02, 0.0001)
    y = amplitude * np.exp(-((x - center) ** 2) / (2 * spread ** 2))
    return x, y


def gaussian_predict(peak, mz):
    x = mz
    center = peak.mz
    amplitude = peak.intensity
    fwhm = peak.full_width_at_half_max
    spread = fwhm / 2.35482
    y = amplitude * np.exp(-((x - center) ** 2) / (2 * spread ** 2))
    return y


def gaussian_error(peak, mz, intensity):
    y = gaussian_predict(peak, mz)
    return intensity - y


def gaussian_volume(peak):
    x, y = gaussian_shape(peak)
    return np.trapz(y, x, dx=0.0001)


def lorentzian_predict(peak, mz):
    center = peak.mz
    fwhm = peak.full_width_at_half_max
    x = mz
    spread = fwhm / 2.
    a = peak.intensity
    b = (spread ** 2)
    c = (x - center) ** 2 + spread ** 2
    return a * (b / c)


def lorentzian_shape(peak):
    center = peak.mz
    fwhm = peak.full_width_at_half_max
    x = np.arange(center - fwhm - 0.02, center + fwhm + 0.02, 0.0001)
    return x, lorentzian_predict(peak, x)


def lorentzian_error(peak, mz, intensity):
    y = lorentzian_predict(peak, mz)
    return intensity - y


def lorentzian_volume(peak):
    x, y = lorentzian_shape(peak)
    return np.trapz(y, x, dx=0.0001)


class PeakShapeModel(object):
    def __init__(self, peak):
        self.peak = peak
        self.center = peak.mz

    def __repr__(self):
        return "{self.__class__.__name__}({self.peak})".format(self=self)


try:
    # Import C extension base class which simply sets up the type
    # signature for later classes to inherit from
    from ms_peak_picker._c.peak_statistics import PeakShapeModel
except ImportError:
    pass


class GaussianModel(PeakShapeModel):
    def shape(self):
        return gaussian_shape(self.peak)

    def predict(self, mz):
        return gaussian_predict(self.peak, mz)

    def volume(self):
        return gaussian_volume(self.peak)

    def error(self, mz, intensity):
        return gaussian_error(self.peak, mz, intensity)


try:
    # Import the accelerated gaussian shape implementation
    from ms_peak_picker._c.peak_statistics import GaussianModel
except ImportError:
    pass


class LorentzianModel(PeakShapeModel):
    def shape(self):
        return lorentzian_shape(self.peak)

    def predict(self, mz):
        return lorentzian_predict(self.peak, mz)

    def volume(self):
        return lorentzian_volume(self.peak)

    def error(self, mz, intensity):
        return lorentzian_error(self.peak, mz, intensity)


def peak_area(mz_array, intensity_array, start, stop):
    area = 0.

    for i in range(start + 1, stop):
        x1 = mz_array[i - 1]
        y1 = intensity_array[i - 1]
        x2 = mz_array[i]
        y2 = intensity_array[i]
        area += (y1 * (x2 - x1)) + ((y2 - y1) * (x2 - x1) / 2.)

    return area


try:
    from ._c import peak_statistics as cpeak_statistics
    _find_signal_to_noise = find_signal_to_noise
    _find_full_width_at_half_max = find_full_width_at_half_max
    _find_left_width = find_left_width
    _find_right_width = find_right_width
    _peak_area = peak_area
    _quadratic_fit = quadratic_fit
    _lorentzian_fit = lorentzian_fit

    find_signal_to_noise = cpeak_statistics.find_signal_to_noise
    peak_area = cpeak_statistics.peak_area
    find_full_width_at_half_max = cpeak_statistics.find_full_width_at_half_max
    quadratic_fit = cpeak_statistics.quadratic_fit
    lorentzian_fit = cpeak_statistics.lorentzian_fit
    find_left_width = cpeak_statistics.find_left_width
    find_right_width = cpeak_statistics.find_right_width
except ImportError:
    pass



cimport cython
cimport numpy as np

ctypedef np.float64_t DTYPE_t
ctypedef np.float64_t mz_t

cpdef size_t nearest_left(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=*)
cpdef size_t nearest_right(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=*)
cpdef size_t get_nearest_binary(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=*, object stop_index=*)
cpdef size_t get_nearest(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index)



# -*- coding: utf-8 -*-
'''
This module contains a simple translation of Gordon Slyz's implementation of Senko's
Fourier Patterson Charge State determination algorithm.
'''

import numpy as np
from scipy import interpolate


def autocorr(inp):
    """
    Compute the autocorrelation for the input peak sequence

    Parameters
    ----------
    inp : np.ndarray(float64)
        An array representing the sequence of intensities around the peak of interest

    Returns
    -------
    np.ndarray(float64)
        The autocorrelation of the input
    """
    out = np.zeros_like(inp)
    n = len(inp)
    average = inp.mean()

    for i in range(n):
        total = 0.0
        top = n - i - 1
        for j in range(top):
            total += (inp[j] - average) * (inp[j + i] - average)
        if j > 0:
            out[i] = (total / float(n))
        else:
            out[i] = 0.
    return out


def _autocorrelation_scores(peak_index, peak, fwhm, min_mz, max_mz, left_index, right_index):
    """
    Compute an approximation of the fourier transform of the signals around the `peak`. The approximation
    of the fit is performed using the autocorrelation, which is then followed up with cubic spline.

    Parameters
    ----------
    peak_index : PeakIndex
        The signal container
    peak : FittedPeak
        The peak to fit around
    fwhm : float
        The area around the peak to fit
    min_mz : float
        The minimum m/z to search from
    max_mz : float
        The maximum m/z to search to
    left_index : int
        The index of the minimum m/z
    right_index : int
        The index of the maximum m/z

    Raises
    ------
    ValueError
        If not enough points can be found for spline interpolation

    Returns
    -------
    autoscores : np.ndarray(float64)
        The autocorrelation of the signals around `peak`
    num_l: int
        The number of points fitted over

    """
    filtered = peak_index
    sum_of_differences = 0.
    point_count = 0.

    for i in range(len(filtered.mz_array) - 1):
        y1 = filtered.intensity_array[i]
        y2 = filtered.intensity_array[i + 1]
        if y1 > 0 and y2 > 0:
            x1 = filtered.mz_array[i]
            x2 = filtered.mz_array[i + 1]

            sum_of_differences += x2 - x1
            point_count += 1
    if point_count > 5:
        average_differences = sum_of_differences / point_count

        num_l = (np.ceil((max_mz - min_mz) / average_differences))
        num_l = int(num_l + num_l * 1.1)
    else:
        num_points = right_index - left_index + 1
        desired_number = 256
        point_multiplier = np.ceil(desired_number / float(num_points))

        if num_points < 5:
            raise ValueError("Not enough points")

        if num_points < desired_number:
            point_multiplier = max(5, point_multiplier)
            num_l = point_multiplier * num_points
        else:
            num_l = num_points

    interpolant = interpolate.interp1d(filtered.mz_array, filtered.intensity_array, kind='cubic')

    x = np.zeros(num_l)  # [0 for i in range(num_l)]
    y = np.zeros(num_l)

    for i in range(num_l):
        xval = (min_mz + ((max_mz - min_mz) * i) / num_l)
        yval = interpolant(xval)
        x[i] = xval
        y[i] = yval

    autoscores = autocorr(y)
    return autoscores, num_l


def fft_patterson_charge_state(peak_index, peak, max_charge=25):
    """
    Computes the optimal charge state for the given peak using the
    FFT+Patterson Charge State Deconvolution method as published in [1]
    and implemented in Decon2LS.

    A sub-region of the spectrum based upon peak width is selected for
    consideration.

    Parameters
    ----------
    peak_index : PeakIndex
        The PeakIndex which the given Peak is derived from,
        and in which surrounding spectral information will
        be extracted
    peak : FittedPeak
        The peak to compute the optimal charge state for.
    max_charge : int, optional
        The maximal charge state to consider. Defaults to 25.

    Returns
    -------
    charge: positive int
        The optimal charge state for the given peak. In the event
        that the algorithm fails to find an acceptable charge state,
        it will return -1 instead.


    References
    ----------
    [1] Senko, M. W., Beu, S. C., & McLafferty, F. W. (1995).
        Automated assignment of charge states from resolved isotopic peaks for multiply charged ions.
        Journal of the American Society for Mass Spectrometry, 6(1), 52–56.
        http://doi.org/10.1016/1044-0305(94)00091-D

    """
    plus = 1.1
    minus = 0.1
    fwhm = peak.full_width_at_half_max

    left_index = peak_index.get_nearest(peak.mz - fwhm - minus, peak.index)
    right_index = peak_index.get_nearest(peak.mz + fwhm + plus, peak.index)

    filtered = peak_index.slice(left_index, right_index)

    min_mz = filtered.mz_array[0]
    max_mz = filtered.mz_array[len(filtered.mz_array) - 1]

    autoscores, num_l = _autocorrelation_scores(filtered, peak, fwhm, min_mz, max_mz, left_index, right_index)

    start_index = 0
    while (start_index < num_l - 1 and autoscores[start_index] > autoscores[start_index + 1]):
        start_index += 1
    best_autoscore = -1
    best_charge_state = -1

    best_autoscore, best_charge_state = _best_scoring_charge_state(
        min_mz, max_mz, start_index, autoscores, max_charge)

    if best_charge_state == -1:
        return -1

    return_charge_state = -1

    charge_state_list, charge_state_score_map = _generate_charge_state_data(
        min_mz, max_mz, start_index, autoscores, max_charge, best_autoscore)

    for i in range(len(charge_state_list)):
        temp_charge_state = charge_state_list[i]
        skip = False
        for j in range(i):
            if charge_state_list[j] == temp_charge_state:
                skip = True
                break
        if skip:
            continue

        if temp_charge_state > 0:
            another_peak = peak.mz + (1.003 / temp_charge_state)
            if peak_index.has_peak_within_tolerance(another_peak, fwhm):
                return_charge_state = temp_charge_state
                if peak.mz * temp_charge_state < 3000:
                    break
                else:
                    return temp_charge_state
    return return_charge_state


def _best_scoring_charge_state(min_mz, max_mz, start_index, autocorr_scores, max_charge):
    best_autoscore = -1
    best_charge_state = -1

    going_up = False
    was_going_up = False

    npoints = len(autocorr_scores)
    for i in range(start_index, npoints):
        if i < 2:
            continue
        going_up = (autocorr_scores[i] - autocorr_scores[i - 1]) > 0
        if was_going_up and not going_up:
            charge_state = int(npoints / ((max_mz - min_mz) * (i - 1) + 0.5))
            current_score = autocorr_scores[i - 1]
            if abs(current_score / autocorr_scores[0]) > 0.05 and charge_state < max_charge:
                if abs(current_score) > best_autoscore:
                    best_autoscore = abs(current_score)
                    best_charge_state = charge_state
        was_going_up = going_up
    return best_autoscore, best_charge_state


def _generate_charge_state_data(min_mz, max_mz, start_index, autocorr_scores,
                                max_charge, best_autoscore):
    charge_state_list = []
    going_up = False
    was_going_up = False

    npoints = len(autocorr_scores)

    charge_state_score_map = dict()

    for i in range(start_index, npoints):
        if i < 2:
            continue
        going_up = autocorr_scores[i] > autocorr_scores[i - 1]

        if was_going_up and not going_up:
            current_charge_state = (npoints / ((max_mz - min_mz) * (i - 1)))

            current_autocorr_score = autocorr_scores[i - 1]
            if ((current_autocorr_score > best_autoscore * 0.1) and current_charge_state < max_charge):
                charge_state = int(round(current_charge_state))
                charge_state_score_map[charge_state] = current_autocorr_score
                charge_state_list.append(charge_state)

        was_going_up = going_up
    return charge_state_list, charge_state_score_map


try:
    _p_autocorr = autocorr
    from ._c.fft_patterson_charge_state import autocorr, _autocorrelation_scores
except ImportError:
    pass



import numpy as np

from .utils import Base
from .fticr_denoising import denoise as fticr_remove_baseline
from .smoothing import gaussian_smooth

try:
    basestring
except NameError:
    from six import string_types as basestring


#: Global register of all named scan filters
filter_register = {}


def register(name, *args, **kwargs):
    """Decorate a class to register a name for
    it, optionally with a set of associated initialization
    parameters.

    Parameters
    ----------
    name : str
        The name to register the filter under.
    *args
        Positional arguments forwarded to the decorated class's
        initialization method
    **kwargs
        Keyword arguments forwarded to the decorated class's
        intialization method

    Returns
    -------
    function
        A decorating function which will carry out the registration
        process on the decorated class
    """
    def wrap(cls):
        filter_register[name] = cls(*args, **kwargs)
        return cls
    return wrap


class FilterBase(Base):
    """A base type for Filters over raw signal arrays.

    All subtypes should provide a :meth:`filter` method
    which takes arguments *mz_array* and *intensity_array*
    which will be NumPy Arrays.
    """
    def filter(self, mz_array, intensity_array):
        return mz_array, intensity_array

    def __call__(self, mz_array, intensity_array):
        return self.filter(mz_array, intensity_array)


@register("median")
class MedianIntensityFilter(FilterBase):
    """Filter signal below the median signal
    """
    def filter(self, mz_array, intensity_array):
        mask = intensity_array < np.median(intensity_array)
        intensity_array = np.array(intensity_array)
        intensity_array[mask] = 0.
        return mz_array, intensity_array


@register("mean_below_mean")
class MeanBelowMeanFilter(FilterBase):
    """Filter signal below the mean below the mean
    """
    def filter(self, mz_array, intensity_array):
        mean = intensity_array.mean()
        mean_below_mean = (intensity_array < mean).mean()
        mask = intensity_array < mean_below_mean
        intensity_array[mask] = 0.
        return mz_array, intensity_array


@register("savitsky_golay")
class SavitskyGolayFilter(FilterBase):
    """Apply `Savitsky-Golay smoothing <https://en.wikipedia.org/wiki/Savitzky%E2%80%93Golay_filter>`_
    to the signal.

    Attributes
    ----------
    deriv : int
        Number of derivatives to take
    polyorder : int
        Order of the polynomial to construct
    window_length : int
        Number of data points to include around the current point
    """
    def __init__(self, window_length=5, polyorder=3, deriv=0):
        self.window_length = window_length
        self.polyorder = polyorder
        self.deriv = deriv

    def filter(self, mz_array, intensity_array):
        from scipy.signal import savgol_filter
        smoothed = savgol_filter(
            intensity_array, window_length=self.window_length,
            polyorder=self.polyorder, deriv=self.deriv).clip(0)
        mask = smoothed > 0
        smoothed[~mask] = 0
        return mz_array, smoothed


@register("gaussian_smooth", 0.02)
class GaussianSmoothFilter(FilterBase):
    def __init__(self, width=0.02):
        self.width = 0.02

    def filter(self, mz_array, intensity_array):
        intensity_array = gaussian_smooth(mz_array, intensity_array, self.width)
        return mz_array, intensity_array


@register("tenth_percent_of_max")
@register("one_percent_of_max", p=0.01)
class NPercentOfMaxFilter(FilterBase):
    def __init__(self, p=0.001):
        self.p = p

    def filter(self, mz_array, intensity_array):
        mask = (intensity_array / intensity_array.max()) < self.p
        intensity_array_clone = np.array(intensity_array)
        intensity_array_clone[mask] = 0.
        return mz_array, intensity_array_clone


@register("fticr_baseline")
class FTICRBaselineRemoval(FilterBase):
    """Apply FTICR baseline removal.

    This calls :py:func:`~ms_peak_picker.fticr_denoising.denoise`

    Attributes
    ----------
    region_width : float
        The width of the region to group windows
        by.
    scale : float
        The multiplier of the noise level to remove.
    window_length : float
        The size of the window to tile across each
        region

    See Also
    --------
    ms_peak_picker.fticr_denoising.denoise
    """
    def __init__(self, window_length=1., region_width=10, scale=5):
        self.window_length = window_length
        self.region_width = region_width
        self.scale = scale

    def filter(self, mz_array, intensity_array):
        return fticr_remove_baseline(mz_array, intensity_array, self.window_length, self.region_width, self.scale)


@register("linear_resampling", 0.01)
class LinearResampling(FilterBase):
    def __init__(self, spacing):
        self.spacing = spacing

    def filter(self, mz_array, intensity_array):
        lo = mz_array.min()
        hi = mz_array.max()
        new_mz = np.arange(lo, hi + self.spacing, self.spacing)
        new_intensity = np.interp(new_mz, mz_array, intensity_array)
        return new_mz, new_intensity


@register("over_10", 10)
@register("over_100", 100)
class ConstantThreshold(FilterBase):
    def __init__(self, constant):
        self.constant = constant

    def filter(self, mz_array, intensity_array):
        mask = intensity_array < self.constant
        intensity_array = intensity_array.copy()
        intensity_array[mask] = 0
        return mz_array, intensity_array


@register("extreme_scale_limiter", 30e3)
class MaximumScaler(FilterBase):
    def __init__(self, threshold):
        self.threshold = threshold

    def filter(self, mz_array, intensity_array):
        if intensity_array.max() > self.threshold:
            intensity_array = intensity_array / intensity_array.max() * self.threshold
        return mz_array, intensity_array


@register("magnitude_boost", 100.0)
class IntensityScaler(FilterBase):
    def __init__(self, scale):
        self.scale = scale

    def filter(self, mz_array, intensity_array):
        intensity_array = intensity_array * self.scale
        return mz_array, intensity_array


class RecalibrateMass(FilterBase):
    def __init__(self, offset):
        self.offset = offset

    def filter(self, mz_array, intensity_array):
        return mz_array + self.offset, intensity_array


def transform(mz_array, intensity_array, filters=None):
    """Apply a series of *filters* to the paired m/z and intensity
    arrays.

    The `filters` argument should be an iterable of either strings,
    callables, or instances of :class:`FilterBase`-derived classes.
    If they are strings, they must be registered names, as created by
    :func:`register`.

    Parameters
    ----------
    mz_array : np.ndarray[float64]
        The m/z array to filter
    intensity_array : np.ndarray[float64]
        The intensity array to filter
    filters : Iterable
        An Iterable of either strings, callables, or instances
        of :class:`FilterBase`-derived classes. If they are
        strings, they must be registered names, as created by
        :func:`register`

    Returns
    -------
    np.ndarray[float64]:
        The m/z array after filtering
    np.ndarray[float64]:
        The intensity array after filtering
    """
    if filters is None:
        filters = []

    for filt in filters:
        if isinstance(filt, basestring):
            filt = filter_register[filt]
        mz_array, intensity_array = filt(mz_array, intensity_array)

    return mz_array, intensity_array


def cast(arg):
    try:
        return int(arg)
    except ValueError:
        try:
            return float(arg)
        except ValueError:
            return str(arg)


def parse(text):
    tokens = text.split(" ")
    filter_name = tokens[0]
    params = list(map(cast, tokens[1:]))
    filter_tp = None
    if filter_name in filter_register:
        filter_instance = filter_register[filter_name]
        if len(tokens) == 1:
            return filter_instance
        else:
            filter_tp = filter_instance.__class__
    else:
        filter_name = filter_name.lower()
        for tp in FilterBase.__subclasses__():
            if tp.__name__.lower() == filter_name:
                filter_tp = tp
                break
    if filter_tp is None:
        raise KeyError(filter_name)
    else:
        return filter_tp(*params)



# cython: embedsignature=True

cimport numpy as np
cimport cython

from libc.math cimport fabs
# ctypedef cython.floating mz_t

@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef size_t nearest_left(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=0):
    """Locate the index nearest to `target_val` in `vec` searching
    to the left of `start_index`
    
    Parameters
    ----------
    vec : np.array
        The array to search
    target_val : float
        The value to search for
    start_index : int, optional
        The starting point to search from
    
    Returns
    -------
    int
        The nearest index
    """
    cdef:
        size_t nearest_index, next_index
        DTYPE_t next_val, best_distance, dist
    nearest_index = start_index
    next_index = start_index

    if next_index == 0:
        return 0
    next_val = vec[next_index]
    best_distance = fabs(target_val - next_val)
    while next_val > target_val:
        next_index -= 1
        next_val = vec[next_index]
        dist = fabs(next_val - target_val)
        if dist < best_distance:
            best_distance = dist
            nearest_index = next_index
        if next_index == 0:
            break
    return nearest_index


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef size_t nearest_right(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=0):
    """Locate the index nearest to `target_val` in `vec` searching
    to the right of `start_index`
    
    Parameters
    ----------
    vec : np.array
        The array to search
    target_val : float
        The value to search for
    start_index : int, optional
        The starting point to search from
    
    Returns
    -------
    int
        The nearest index
    """
    cdef:
        size_t nearest_index, next_index, size
        DTYPE_t next_val, best_distance, dist
    nearest_index = start_index
    next_index = start_index

    size = vec.shape[0] - 1
    if next_index >= size:
        return size

    next_val = vec[next_index]
    best_distance = fabs(next_val - target_val)
    while (next_val < target_val):
        next_index += 1
        next_val = vec[next_index]
        dist = fabs(next_val - target_val)
        if dist < best_distance:
            best_distance = dist
            nearest_index = next_index
        if next_index == size:
            break
    return nearest_index


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef size_t get_nearest_binary(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index=0, object stop_index=None):
    """Find the nearest index to `target_val` in `vec` using binary
    search
    
    Parameters
    ----------
    vec : np.ndarray
        The array to search
    target_val : float
        The value to search for
    start_index : int, optional
        The lower bound index
    stop_index : None, optional
        The upper bound index
    
    Returns
    -------
    int
        The nearest index
    """
    cdef:
        DTYPE_t min_val, max_val, ratio, mid_next_val, mid_val, cval
        size_t stop_index_, mid_index

    if stop_index is None:
        stop_index_ = vec.shape[0] - 1
    else:
        stop_index_ = stop_index
    cval = vec[start_index]
    if cval > target_val:
        return start_index

    cval = vec[stop_index_]
    if cval < target_val:
        return stop_index_
    while True:
        min_val = vec[start_index]
        max_val = vec[stop_index_]
        if fabs(stop_index_ - start_index) <= 1 and (target_val >= min_val) and (target_val <= max_val):
            if fabs(min_val - target_val) < fabs(max_val - target_val):
                return start_index
            return stop_index_
        ratio = (max_val - target_val) / (max_val - min_val)
        mid_index = <size_t>(start_index * ratio + stop_index_ * (1 - ratio) + 0.5)
        if (mid_index == start_index):
            mid_index = start_index + 1
        elif (mid_index == stop_index_):
            mid_index = stop_index_ - 1

        mid_val = vec[mid_index]

        if mid_val >= target_val:
            stop_index_ = mid_index
        elif mid_index + 1 == stop_index_:
            if fabs(mid_val - target_val) < fabs(max_val - target_val):
                return mid_index
            return stop_index_
        else:
            mid_next_val = vec[mid_index + 1]
            if target_val >= mid_val and target_val <= mid_next_val:
                if (target_val - mid_val < mid_next_val - mid_val):
                    return mid_index
                return mid_index + 1
            start_index = mid_index + 1


@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cpdef size_t get_nearest(np.ndarray[mz_t, ndim=1, mode='c'] vec, DTYPE_t target_val, size_t start_index):
    """Locate the index nearest to `target_val` in `vec`. Starts
    by interpolating from the value at `start_index` to `target_val`
    and then starts searching to the left and right of the guessed
    index.

    Parameters
    ----------
    vec : np.array
        The array to search
    target_val : float
        The value to search for
    start_index : int, optional
        The starting point to search from
    
    Returns
    -------
    int
        The nearest index
    """
    cdef:
        size_t size, next_index, move_by
        DTYPE_t step, next_val
        DTYPE_t distance_remaining


    size = (vec.shape[0]) - 1
    if target_val >= vec[size]:
        return size
    elif target_val <= vec[0]:
        return 0

    distance_remaining = target_val - vec[start_index]

    if start_index < size:
        step = vec[start_index + 1] - vec[start_index]
    else:
        step = vec[start_index] - vec[start_index - 1]

    move_by = <size_t>(distance_remaining / step)

    next_index = start_index + move_by

    if next_index < 0:
        next_index = 0
    if next_index > size:
        next_index = size - 1

    next_val = vec[next_index]
    if target_val >= next_val:
        return nearest_right(vec, target_val, next_index)
    else:
        return nearest_left(vec, target_val, next_index)



import numpy as np
from .search import get_nearest
from .fft_patterson_charge_state import fft_patterson_charge_state
from .peak_statistics import peak_area


class PeakIndex(object):
    """Represent a pair of m/z and intensity arrays paired with
    their :class:`PeakSet`.

    This class provides wrappers for most :class:`PeakSet` methods which
    it provides for its :attr:`peaks` attribute, and relates these peaks
    to their raw signal, making re-integration more convenient.

    This type also provides some limited multi-peak analysis
    to estimate the charge state of a peak using the isotopic pattern-free
    using Senko's Fourier Patterson Charge State determination algorithm
    :meth:`fft_patterson_charge_state`

    Attributes
    ----------
    mz_array : np.ndarray
        The original m/z array the peaks were picked from
    intensity_array : np.ndarray
        The original intensity array the peaks were picked from
    peaks : PeakSet
        The set of :class:`FittedPeak` objects picked from the associated
        arrays
    """

    def __init__(self, mz_array, intensity_array, peaks):
        self.mz_array = mz_array
        self.intensity_array = intensity_array
        self.peaks = peaks

    def pack(self):
        """Create a copy of `self` with the large
        arrays stripped out. This removes most functionality
        beyond wrapping the underlying :class:`PeakSet` but
        makes the object smaller.

        Returns
        -------
        PeakIndex
        """
        return PeakIndex(np.array([], dtype=np.float64), np.array([], dtype=np.float64), self.peaks.clone())

    def clone(self):
        """Create a deep copy of `self`

        Returns
        -------
        PeakIndex
        """
        return PeakIndex(np.array(self.mz_array), np.array(self.intensity_array), self.peaks.clone())

    def get_nearest(self, mz, index):
        """Get the nearest index to `mz` in the underlying arrays

        Parameters
        ----------
        mz : float
            The m/z to search for
        index : int
            The index to search from

        Returns
        -------
        int

        Raises
        ------
        ValueError
            If the underlying arrays have been stripped, this method
            cannot be used
        """
        if self.mz_array is None or len(self.mz_array) == 0:
            raise ValueError(
                "Cannot call get_nearest when raw arrays are None")

        return get_nearest(self.mz_array, mz, index)

    def get_nearest_peak(self, mz):
        """Wraps :meth:`PeakSet.get_nearest_peak`

        Parameters
        ----------
        mz : float
            The m/z to search with

        Returns
        -------
        tuple of (FittedPeak, float)
            The nearest peak and the m/z delta between that peak and `mz`
        """
        return self.peaks.get_nearest_peak(mz)

    def slice(self, start, stop):
        return (PeakIndex(self.mz_array[start:stop], self.intensity_array[start:stop],
                          self.peaks.between(self.mz_array[start], self.mz_array[stop] + 1)))

    def between(self, start, stop):
        """Wraps :meth:`PeakSet.between`

        Parameters
        ----------
        start : float
            The lower m/z limit
        stop : float
            The upper m/z limit

        Returns
        -------
        PeakSet
        """
        return self.peaks.between(start, stop)

    def all_peaks_for(self, mz, error_tolerance=2e-5):
        """Find all peaks within `error_tolerance` ppm of `mz` m/z.

        Parameters
        ----------
        mz : float
            The query m/z
        error_tolerance : float, optional
            The parts-per-million error tolerance (the default is 2e-5)

        Returns
        -------
        tuple
        """
        return self.peaks.all_peaks_for(mz, error_tolerance)

    def has_peak_within_tolerance(self, mz, tol):
        return has_peak_within_tolerance(self.peaks, mz, tol)

    def __getitem__(self, index):
        return self.peaks[index]

    def __eq__(self, other):
        return self.peaks == other.peaks

    def __ne__(self, other):
        return self.peaks != other.peaks

    def has_peak(self, mz, tolerance=2e-5):
        """Wraps :meth:`PeakSet.has_peak`

        Parameters
        ----------
        mz : float
            The m/z to search for
        tolerance : float, optional
            The error tolerance to accept. Defaults to 2e-5 (20 ppm)

        Returns
        -------
        FittedPeak
            The peak, if found. `None` otherwise.
        """
        return self.peaks.has_peak(mz, tolerance)

    predict_charge_state = fft_patterson_charge_state

    def __len__(self):
        return len(self.peaks)

    def area(self, peak):
        if self.mz_array is None:
            raise ValueError("Cannot call area when raw arrays are None")

        lo = self.get_nearest(
            peak.mz - peak.full_width_at_half_max, peak.index)
        hi = self.get_nearest(
            peak.mz + peak.full_width_at_half_max, peak.index)
        return peak_area(self.mz_array, self.intensity_array, lo, hi)

    def points_along(self, peak, width=None):
        if self.mz_array is None:
            raise ValueError(
                "Cannot call points_along when raw arrays are None")

        if width is None:
            width = peak.full_width_at_half_max
        lo = self.get_nearest(peak.mz - width, peak.index)
        hi = self.get_nearest(peak.mz + width, peak.index)
        return self.mz_array[lo:hi], self.intensity_array[lo:hi]

    def set_peaks(self, peaks):
        self.peaks = self.peaks.__class__(tuple(peaks))

    def __repr__(self):
        return "PeakIndex(%d points, %d peaks)" % (len(self.mz_array), len(self.peaks))


def has_peak_within_tolerance(peaklist, mz, tol):
    lo = 0
    hi = len(peaklist)

    def sweep(lo, hi):
        for i in range(hi - lo):
            i += lo
            v = peaklist[i].mz
            if abs(v - mz) < tol:
                return i or True
        return False

    def binsearch(lo, hi):
        if (hi - lo) < 5:
            return sweep(lo, hi)
        else:
            mid = (hi + lo) // 2
            v = peaklist[mid].mz
            if abs(v - mz) < tol:
                return mid
            elif v > mz:
                return binsearch(lo, mid)
            else:
                return binsearch(mid, hi)
    return binsearch(lo, hi)


_PeakIndex = PeakIndex
try:
    from ms_peak_picker._c.peak_index import PeakIndex
except:
    pass



import numpy as np


def gauss(x, mean, variance, scale=2):
    return 1. / np.sqrt(2 * np.pi * variance) * np.exp(-(x - mean) ** 2 / (2 * variance))


def binsearch(array, value):
    lo = 0
    hi = len(array)
    while hi != lo:
        mid = (hi + lo) / 2
        point = array[mid]
        if value == point:
            return mid
        elif hi - lo == 1:
            return mid
        elif point > value:
            hi = mid
        else:
            lo = mid


def gaussian_smooth(x, y, width=0.02):
    smoothed = np.zeros_like(y)

    n = x.shape[0]
    for i in range(n):
        low_edge = binsearch(x, x[i] - width)
        high_edge = binsearch(x, x[i] + width)
        if high_edge - 1 == low_edge or high_edge == low_edge:
            smoothed[i] = y[i]
            continue
        x_slice = x[low_edge:high_edge]
        y_slice = y[low_edge:high_edge]
        center = np.mean(x_slice)
        spread = np.var(x_slice)
        weights = gauss(x_slice, center, spread)
        val = ((y_slice * weights).sum() / weights.sum())
        if np.isnan(val):
            raise ValueError("NaN")
        smoothed[i] = val
    return smoothed


try:
    _has_c = True
    _gaussian_smooth = gaussian_smooth
    from ms_peak_picker._c.smoother import gaussian_smooth
except ImportError:
    _has_c = False



# cython: embedsignature=True
cimport cython
import operator
from libc.stdlib cimport malloc, realloc, free

from libc.stdio cimport printf

from cpython.tuple cimport PyTuple_GET_ITEM, PyTuple_GetItem, PyTuple_GetSlice, PyTuple_GET_SIZE
from cpython cimport PyObject

cdef double ppm_error(double x, double y):
    return (x - y) / y


cdef double INF = float('inf')


@cython.freelist(1000000)
cdef class PeakBase:

    cpdef PeakBase clone(self):
        return PeakBase()

    cpdef PeakBase copy(self):
        return self.clone()


cdef class FittedPeak(PeakBase):
    """Represent a single centroided mass spectral peak.

    FittedPeak instances are comparable for equality and
    hashable.

    Attributes
    ----------
    mz : float
        The m/z value at which the peak achieves its maximal
        abundance
    intensity : float
        The apex height of the peak
    signal_to_noise : float
        The signal to noise ratio of the peak
    full_width_at_half_max : double
        The symmetric full width at half of the
        peak's maximum height
    index : int
        The index at which the peak was found in the m/z array
        it was picked from
    peak_count : int
        The order in which the peak was picked
    area : float
        The total area of the peak, as determined
        by trapezoid integration
    left_width : float
        The left-sided width at half of max
    right_width : float
        The right-sided width at half of max
    """
    def __init__(self, mz, intensity, signal_to_noise, peak_count, index, full_width_at_half_max,
                 area, left_width=0, right_width=0):
            self.mz = mz
            self.intensity = intensity
            self.signal_to_noise = signal_to_noise
            self.peak_count = peak_count
            self.index = index
            self.full_width_at_half_max = full_width_at_half_max
            self.area = area
            self.left_width = left_width
            self.right_width = right_width

    @staticmethod
    cdef FittedPeak _create(double mz, double intensity, double signal_to_noise,
                            double full_width_at_half_max, double left_width,
                            double right_width, long peak_count, long index,
                            double area):
        cdef:
            FittedPeak inst

        inst = FittedPeak.__new__(FittedPeak)
        inst.mz = mz
        inst.intensity = intensity
        inst.signal_to_noise = signal_to_noise
        inst.full_width_at_half_max = full_width_at_half_max
        inst.left_width = left_width
        inst.right_width = right_width
        inst.peak_count = peak_count
        inst.index = index
        inst.area = area
        return inst

    cpdef PeakBase clone(self):
        return FittedPeak._create(
            self.mz, self.intensity, self.signal_to_noise,
            self.full_width_at_half_max, self.left_width, self.right_width,
            self.peak_count, self.index, self.area)

    def __repr__(self):
        return ("FittedPeak(mz=%0.3f, intensity=%0.3f, signal_to_noise=%0.3f, peak_count=%d, "
                "index=%d, full_width_at_half_max=%0.3f, area=%0.3f)") % (
                self.mz, self.intensity, self.signal_to_noise,
                self.peak_count, self.index, self.full_width_at_half_max,
                self.area)

    def __getstate__(self):
        return (self.mz, self.intensity, self.signal_to_noise,
                self.peak_count, self.index, self.full_width_at_half_max,
                self.area, self.left_width, self.right_width)

    def __setstate__(self, state):
        (self.mz, self.intensity, self.signal_to_noise,
         self.peak_count, self.index, self.full_width_at_half_max,
         self.area, self.left_width, self.right_width) = state

    def __reduce__(self):
        return FittedPeak, (self.mz, self.intensity, self.signal_to_noise, self.peak_count,
                            self.index, self.full_width_at_half_max, self.left_width,
                            self.right_width)

    def __hash__(self):
        return hash((self.mz, self.intensity, self.signal_to_noise, self.full_width_at_half_max))

    cpdef bint _eq(self, FittedPeak other):
        return (abs(self.mz - other.mz) < 1e-5) and (
            abs(self.intensity - other.intensity) < 1e-5) and (
            abs(self.signal_to_noise - other.signal_to_noise) < 1e-5) and (
            abs(self.full_width_at_half_max - other.full_width_at_half_max) < 1e-5)

    def __richcmp__(self, other, int code):
        if code == 2:
            if other is None:
                return False
            return self._eq(other)
        elif code == 3:
            if other is None:
                return True
            return not self._eq(other)



cdef class PeakSet(object):
    """A sequence of :class:`FittedPeak` instances, ordered by m/z,
    providing efficient search and retrieval of individual peaks or
    whole intervals of the m/z domain.

    This collection is not meant to be updated once created, as it
    it is indexed for ease of connecting individual peak objects to
    their position in the underlying sequence.

    One :class:`PeakSet` is considered equal to another if all of their
    contained :class:`FittedPeak` members are equal to each other

    Attributes
    ----------
    peaks : tuple
        The :class:`FittedPeak` instances, stored
    """

    @staticmethod
    cdef PeakSet _create(tuple peaks):
        cdef:
            PeakSet inst
        inst = PeakSet.__new__(PeakSet)
        inst.peaks = peaks
        return inst

    def __init__(self, peaks):
        self.peaks = tuple(peaks)

    def __len__(self):
        return PyTuple_GET_SIZE(self.peaks)

    cdef size_t get_size(self):
        return PyTuple_GET_SIZE(self.peaks)

    cpdef reindex(self):
        """Re-indexes the sequence of peaks, updating their
        :attr:`peak_count` and setting their :attr:`index` if
        it is missing.
        """
        self._index()

    def _index(self):
        self.peaks = tuple(sorted(self.peaks, key=operator.attrgetter('mz')))
        i = 0
        for i, peak in enumerate(self.peaks):
            peak.peak_count = i
            if peak.index == -1:
                peak.index = i
        self.indexed = True
        return i

    def has_peak(self, double mz, double tolerance=1e-5):
        """Search the for the peak nearest to `mz` within
        `tolerance` error (in PPM)

        Parameters
        ----------
        mz : float
            The m/z to search for
        tolerance : float, optional
            The error tolerance to accept. Defaults to 1e-5 (10 ppm)

        Returns
        -------
        FittedPeak
            The peak, if found. `None` otherwise.
        """
        peak = self._has_peak(mz, tolerance)
        if peak is _null_peak:
            return None
        return peak

    def get_nearest_peak(self, double mz):
        cdef:
            FittedPeak peak
            double err
        peak = self._get_nearest_peak(mz, &err)
        return peak, err

    cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):
        cdef:
            FittedPeak peak
        peak = _binary_search_nearest_match(self.peaks, mz, 0, PyTuple_GET_SIZE(self.peaks), errout)
        return peak

    cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):
        cdef:
            FittedPeak peak
        peak = binary_search_ppm_error(self.peaks, mz, tolerance)
        return peak

    cdef FittedPeak getitem(self, size_t i):
        return <FittedPeak>PyTuple_GET_ITEM(self.peaks, i)

    def __repr__(self):
        return "<PeakSet %d Peaks>" % (len(self))

    def __getitem__(self, item):
        if isinstance(item, slice):
            return PeakSet(self.peaks[item])
        return self.peaks[item]

    cpdef PeakSet clone(self):
        """Creates a deep copy of the sequence of peaks

        Returns
        -------
        PeakSet
        """
        cdef PeakSet inst = PeakSet([p.clone() for p in self])
        inst.indexed = self.indexed
        return inst

    cpdef PeakSet copy(self):
        return self.clone()

    def between(self, double m1, double m2):
        """Retrieve a :class:`PeakSet` containing all the peaks
        in `self` whose m/z is between `m1` and `m2`.

        These peaks are not copied.

        Parameters
        ----------
        m1 : float
            The lower m/z limit
        m2 : float
            The upper m/z limit

        Returns
        -------
        PeakSet
        """
        return self._between(m1, m2)

    cdef PeakSet _between(self, double m1, double m2):
        cdef:
            FittedPeak p1
            FittedPeak p2
            double err
            tuple sliced
            size_t start, end, n
        p1 = self._get_nearest_peak(m1, &err)
        p2 = self._get_nearest_peak(m2, &err)
        start = p1.peak_count
        end = p2.peak_count + 1
        n = self.get_size()
        if p1.mz < m1 and start + 1 < n:
            start += 1
        if p2.mz > m2 and end > 0:
            end -= 1
        sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
        return PeakSet._create(sliced)

    cdef int _between_bounds(self, double m1, double m2, size_t* startp, size_t* endp):
        cdef:
            FittedPeak p1
            FittedPeak p2
            double err
            size_t start, end, n

        p1 = self._get_nearest_peak(m1, &err)
        p2 = self._get_nearest_peak(m2, &err)
        start = p1.peak_count
        end = p2.peak_count + 1
        n = self.get_size()
        if p1.mz < m1 and start + 1 < n:
            start += 1
        if p2.mz > m2 and end > 0:
            end -= 1
        startp[0] = start
        endp[0] = end
        return 0

    cpdef tuple all_peaks_for(self, double mz, double error_tolerance=2e-5):
        """Find all peaks within `error_tolerance` ppm of `mz` m/z.

        Parameters
        ----------
        mz : float
            The query m/z
        error_tolerance : float, optional
            The parts-per-million error tolerance (the default is 2e-5)

        Returns
        -------
        tuple
        """
        cdef:
            double lo, hi
            size_t start, end
            tuple sliced

        lo = mz - (mz * error_tolerance)
        hi = mz + (mz * error_tolerance)
        self._between_bounds(lo, hi, &start, &end)
        sliced = <tuple>PyTuple_GetSlice(self.peaks, start, end)
        return sliced

    def __getstate__(self):
        return self.peaks

    def __setstate__(self, state):
        self.peaks = state
        self.indexed = False
        self.reindex()

    def __reduce__(self):
        return PeakSet, (tuple(),), self.__getstate__()

    cpdef bint _eq(self, PeakSet other):
        cdef:
            size_t i, n
            FittedPeak p1, p2

        n = self.get_size()
        if n != other.get_size():
            return False
        for i in range(n):
            p1 = self.getitem(i)
            p2 = other.getitem(i)
            if not p1._eq(p2):
                return False
        return True

    def __richcmp__(self, object other, int code):
        cdef:
            tuple other_tuple

        if other is None:
            if code == 2:
                return False
            elif code == 3:
                return True
        if isinstance(other, PeakSet):
            if code == 2:
                return self._eq(<PeakSet>other)
            elif code == 3:
                return not self._eq(<PeakSet>other)
            else:
                return NotImplemented
        else:
            try:
                other_tuple = tuple(other)
                if code == 2:
                    return self.peaks == other_tuple
                elif code == 3:
                    return self.peaks != other_tuple
                else:
                    return NotImplemented
            except Exception:
                if code == 2:
                    return False
                elif code == 3:
                    return True
                else:
                    return NotImplemented


cdef FittedPeak _null_peak
_null_peak = FittedPeak(-1,-1,-1,-1,-1,-1, -1)


cpdef FittedPeak binary_search_ppm_error(tuple array, double value, double tolerance):
    return _binary_search_ppm_error(array, value, 0, len(array), tolerance)


cdef FittedPeak _binary_search_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):
    cdef:
        size_t mid, lower_edge
        FittedPeak target
        double target_value, error
    if (hi - lo) < 5:
        return _sweep_solution_ppm_error(array, value, lo, hi, tolerance)
    else:
        mid = (hi + lo) // 2
        target = <FittedPeak>PyTuple_GetItem(array, mid)
        target_value = target.mz
        error = ppm_error(value, target_value)
        if abs(error) <= tolerance:
            return _sweep_solution_ppm_error(array, value, max(mid - (mid if mid < 5 else 5), lo), min(mid + 5, hi), tolerance)
        elif target_value > value:
            return _binary_search_ppm_error(array, value, lo, mid, tolerance)
        elif target_value < value:
            return _binary_search_ppm_error(array, value, mid, hi, tolerance)
    return _null_peak


cdef FittedPeak _sweep_solution_ppm_error(tuple array, double value, size_t lo, size_t hi, double tolerance):
    cdef:
        double best_error, error, abs_error
        size_t i, limit

    best_index = -1
    best_intensity = 0
    best_error = 1000000000000000
    for i in range(hi - lo):
        target = <FittedPeak>PyTuple_GetItem(array, lo + i)
        error = ppm_error(value, target.mz)
        abs_error = abs(error)
        if abs_error < tolerance and (abs_error < (best_error * 1.1)) and (target.intensity > best_intensity):
            best_index = lo + i
            best_error = abs_error
    if best_index == -1:
        return _null_peak
    else:
        return <FittedPeak>PyTuple_GetItem(array, best_index)


cdef FittedPeak _sweep_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):
    cdef:
        size_t i
        size_t best_index
        double best_error, err
        double v

    best_error = float('inf')
    best_index = -1
    for i in range(hi - lo):
        i += lo
        v = array[i].mz
        err = abs(v - value)
        if err < best_error:
            best_error = err
            best_index = i
    errout[0] = best_error
    return array[best_index]


cdef FittedPeak _binary_search_nearest_match(tuple array, double value, size_t lo, size_t hi, double* errout):
    cdef:
        size_t mid
        double v
        double err

    if (hi - lo) < 5:
        return _sweep_nearest_match(array, value, lo, hi, errout)
    else:
        mid = (hi + lo) // 2
        v = array[mid].mz
        if abs(v - value) < 1.:
            return _sweep_nearest_match(array, value, lo, hi, errout)
        elif v > value:
            return _binary_search_nearest_match(array, value, lo, mid, errout)
        else:
            return _binary_search_nearest_match(array, value, mid, hi, errout)


@cython.cdivision(True)
cdef size_t double_binary_search_ppm(double* array, double value, double tolerance, size_t n):
    cdef:
        size_t lo, hi, mid
        size_t i, best_ix
        double x, err, best_err, abs_err
    lo = 0
    hi = n
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x - value) / value
        if abs(err) < tolerance:
            i = mid
            best_error = err
            best_ix = mid
            while i > 0:
                i -= 1
                x = array[i]
                err = (x - value) / value
                abs_err = abs(err)
                if abs_err > tolerance:
                    break
                elif abs_err < best_error:
                    best_error = abs_err
                    best_ix = i
            i = mid
            while i < n - 1:
                i += 1
                x = array[i]
                err = (x - value) / value
                abs_err = abs(err)
                if abs_err > tolerance:
                    break
                elif abs_err < best_error:
                    best_error = abs_err
                    best_ix = i
            return best_ix
        elif (hi - 1) == lo:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


@cython.cdivision(True)
cdef size_t double_binary_search_nearest_match(double* array, double value, size_t n):
    cdef:
        size_t lo, hi, mid
        size_t i, best_ix
        double x, err, best_err, abs_err
    lo = 0
    hi = n
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = x - value
        if err == 0 or ((hi - 1) == lo):
            i = mid
            best_err = abs(err)
            best_ix = mid
            while i > 0:
                i -= 1
                x = array[i]
                err = (x - value)
                abs_err = abs(err)
                if abs_err > best_err:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            i = mid
            while i < n - 1:
                i += 1
                x = array[i]
                err = (x - value)
                abs_err = abs(err)
                if abs_err > best_err:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            return best_ix
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


@cython.cdivision(True)
cdef size_t double_binary_search_ppm_with_hint(double* array, double value, double tolerance, size_t n, size_t hint):
    cdef:
        size_t lo, hi, mid
        size_t i, best_ix
        double x, err, best_err, abs_err
    lo = hint
    hi = n
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x - value) / value
        abs_err = abs(err)
        if abs_err < tolerance:
            i = mid
            best_err = abs_err
            best_ix = mid
            while i > 0:
                i -= 1
                x = array[i]
                err = (x - value) / value
                abs_err = abs(err)
                if abs_err > tolerance:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            i = mid
            while i < n - 1:
                i += 1
                x = array[i]
                err = (x - value) / value
                abs_err = abs(err)
                if abs_err > tolerance:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            return best_ix
        elif (hi - 1) == lo:
            return mid
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0

@cython.cdivision(True)
cdef size_t double_binary_search_nearest_match_with_hint(double* array, double value, size_t n, size_t hint):
    cdef:
        size_t lo, hi, mid
        size_t i, best_ix
        double x, err, best_err, abs_err
    lo = hint
    hi = n
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = x - value
        if err == 0 or ((hi - 1) == lo):
            i = mid
            best_err = abs(err)
            best_ix = mid
            while i > 0:
                i -= 1
                x = array[i]
                err = (x - value)
                abs_err = abs(err)
                if abs_err > best_err:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            i = mid
            while i < n - 1:
                i += 1
                x = array[i]
                err = (x - value)
                abs_err = abs(err)
                if abs_err > best_err:
                    break
                elif abs_err < best_err:
                    best_err = abs_err
                    best_ix = i
            return best_ix
        elif err > 0:
            hi = mid
        else:
            lo = mid
    return 0


cdef size_t INTERVAL_INDEX_SIZE = 10000

cdef class PeakSetIndexed(PeakSet):

    @staticmethod
    cdef PeakSet _create(tuple peaks):
        cdef:
            PeakSetIndexed inst
        inst = PeakSetIndexed.__new__(PeakSetIndexed)
        inst.peaks = peaks
        inst.mz_index = NULL
        inst.interval_index = NULL
        return inst

    def __init__(self, peaks):
        PeakSet.__init__(self, peaks)
        self.mz_index = NULL
        self.interval_index = NULL

    def __dealloc__(self):
        if self.mz_index != NULL:
            free(self.mz_index)
        if self.interval_index != NULL:
            free_index_list(self.interval_index)

    cpdef PeakSet clone(self):
        cdef PeakSetIndexed inst = PeakSetIndexed(tuple([p.clone() for p in self]))
        if self.indexed:
            inst.reindex()
        return inst

    cpdef _allocate_index(self):
        cdef:
            size_t i, n
            FittedPeak p
            index_list* interval_index
        n = self.get_size()
        if self.mz_index != NULL:
            free(self.mz_index)
            self.mz_index = NULL
        self.mz_index = <double*>malloc(sizeof(double) * n)
        for i in range(n):
            p = self.getitem(i)
            self.mz_index[i] = p.mz

        if n > 2:
            if self.interval_index != NULL:
                free_index_list(self.interval_index)
                self.interval_index = NULL
            interval_index = <index_list*>malloc(sizeof(index_list))
            build_interval_index(self, interval_index, INTERVAL_INDEX_SIZE)
            if check_index(interval_index) != 0:
                free_index_list(interval_index)
            else:
                self.interval_index = interval_index

    def _index(self):
        i = PeakSet._index(self)
        self._allocate_index()
        return i

    @cython.cdivision(True)
    cdef FittedPeak _has_peak(self, double mz, double tolerance=1e-5):
        cdef:
            size_t i, n, s
            FittedPeak peak
        n = self.get_size()
        if n == 0:
            return _null_peak
        if self.interval_index != NULL:
            find_search_interval(self.interval_index, mz, &s, &n)
            i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, n, s)
        else:
            i = double_binary_search_ppm(self.mz_index, mz, tolerance, n)
        peak = self.getitem(i)
        if abs((peak.mz - mz) / mz) < tolerance:
            return peak
        else:
            return _null_peak

    @cython.cdivision(True)
    def get_nearest_peak(self, double mz):
        cdef:
            double errout
            FittedPeak peak
        peak = self._get_nearest_peak(mz, &errout)
        return peak, errout

    @cython.cdivision(True)
    cdef FittedPeak _get_nearest_peak(self, double mz, double* errout):
        cdef:
            size_t i, n, s
            FittedPeak peak
        n = self.get_size()
        if n == 0:
            errout[0] = INF
            return _null_peak
        if self.interval_index != NULL:
            find_search_interval(self.interval_index, mz, &s, &n)
            i = double_binary_search_nearest_match_with_hint(self.mz_index, mz, n, s)
        else:
            i = double_binary_search_nearest_match(self.mz_index, mz, n)
        peak = self.getitem(i)
        errout[0] = (peak.mz - mz) / mz
        return peak

    cpdef FittedPeak has_peak_hinted(self, double mz, double tolerance=2e-5, size_t hint_start=0, size_t hint_end=-1):
        cdef:
            size_t i, n
            FittedPeak peak
        n = self.get_size()
        if hint_start > n :
            raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_start, n))
        if hint_end > n:
            raise ValueError("Hinted start index %d cannot be greater than the set size %d" % (hint_end, n))
        if n == 0:
            return _null_peak
        i = double_binary_search_ppm_with_hint(self.mz_index, mz, tolerance, hint_end, hint_start)
        peak = self.getitem(i)
        if abs((peak.mz - mz) / mz) < tolerance:
            return peak
        else:
            return _null_peak

    def test_interval(self, double value):
        cdef:
            int status
            size_t start, end
            size_t i

        status = find_search_interval(self.interval_index, value, &start, &end)
        return start, end, interpolate_index(self.interval_index, value)

    def find_interval_for(self, double value):
        return interpolate_index(self.interval_index, value)

    def check_interval(self, size_t i):
        cdef:
            index_cell cell
        cell = self.interval_index.index[i]
        return cell


cdef double* build_linear_spaced_array(double low, double high, size_t n):
    cdef:
        double* array
        double delta
        size_t i

    delta = (high - low) / (n - 1)

    array = <double*>malloc(sizeof(double) * n)

    for i in range(n):
        array[i] = low + i * delta
    return array


cdef void free_index_list(index_list* index):
    free(index.index)
    free(index)


cdef int check_index(index_list* index) nogil:
    if index.size == 0:
        return 1
    elif (index.high - index.low) == 0:
        return 2
    else:
        return 0


cdef int build_interval_index(PeakSet peaks, index_list* index, size_t index_size):
    cdef:
        double* linear_spacing
        double current_value, err, next_value
        size_t i, start_i, end_i, index_i, peaks_size
        FittedPeak peak
    peaks_size = peaks.get_size()
    if peaks_size > 0:
        index.low = peaks.getitem(0).mz
        index.high = peaks.getitem(peaks_size - 1).mz
    else:
        index.low = 0
        index.high = 1
    index.size = index_size
    linear_spacing = build_linear_spaced_array(
        index.low,
        index.high,
        index_size)

    index.index = <index_cell*>malloc(sizeof(index_cell) * index_size)

    for index_i in range(index_size):
        current_value = linear_spacing[index_i]
        peak = peaks._get_nearest_peak(current_value, &err)
        if peaks_size > 0:
            start_i = peak.peak_count
            if index_i > 0:
                start_i = index.index[index_i - 1].end - 1
            if index_i == index_size - 1:
                end_i = peaks_size - 1
            else:
                next_value = linear_spacing[index_i + 1]
                i = peak.peak_count
                while i < peaks_size:
                    peak = peaks.getitem(i)
                    if abs(current_value - peak.mz) > abs(next_value - peak.mz):
                        break
                    i += 1
                end_i = i
        else:
            start_i = 0
            end_i = 0
        index.index[index_i].center_value = current_value
        index.index[index_i].start = start_i
        index.index[index_i].end = end_i

    free(linear_spacing)
    return 0


cdef size_t interpolate_index(index_list* index, double value):
    cdef:
        double v
        size_t i
    v = (((value - index.low) / (index.high - index.low)) * index.size)
    i = <size_t>v
    return i


cdef int find_search_interval(index_list* index, double value, size_t* start, size_t* end):
    cdef:
        size_t i
    if value > index.high:
        i = index.size - 1
    elif value < index.low:
        i = 0
    else:
        i = interpolate_index(index, value)
    if i > 0:
        if i < index.size:
            start[0] = index.index[i - 1].start
        else:
            # if we're at index.size or greater, act as if we're at the last
            # cell of the index
            start[0] = index.index[index.size - 2].start
    else:
        # if somehow the index were negative, this could end badly.
        start[0] = index.index[i].start
    if i >= (index.size - 1):
        end[0] = index.index[index.size - 1].end + 1
    else:
        end[0] = index.index[i + 1].end + 1
    return 0



import sys
import traceback
import os
import platform

from setuptools import setup, Extension, find_packages

from distutils.command.build_ext import build_ext
from distutils.errors import (CCompilerError, DistutilsExecError,
                              DistutilsPlatformError)


def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


# with_openmp = has_option('with-openmp')
no_openmp = has_option('no-openmp')

with_openmp = not no_openmp

use_python_implementation = has_option("pure-python")

print("Building with OpenMP? %s" % with_openmp)


def configure_openmp(ext):
    # http://www.microsoft.com/en-us/download/confirmation.aspx?id=2092 was required.
    if os.name == 'nt' and with_openmp:
        ext.extra_compile_args.append("/openmp")
    elif platform.system() == 'Darwin':
        pass
    elif with_openmp:
        ext.extra_compile_args.append("-fopenmp")
        ext.extra_link_args.append("-fopenmp")


def OpenMPExtension(*args, **kwargs):
    ext = Extension(*args, **kwargs)
    configure_openmp(ext)
    return ext


def make_cextensions():
    import numpy
    try:
        from Cython.Build import cythonize
        extensions = cythonize([
            OpenMPExtension(
                name="ms_peak_picker._c.peak_statistics", sources=["ms_peak_picker/_c/peak_statistics.pyx"],
                include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.peak_set', sources=["ms_peak_picker/_c/peak_set.pyx"]),
            Extension(name='ms_peak_picker._c.fft_patterson_charge_state',
                      sources=["ms_peak_picker/_c/fft_patterson_charge_state.pyx"],
                      include_dirs=[numpy.get_include()]),
            Extension(name="ms_peak_picker._c.search", sources=["ms_peak_picker/_c/search.pyx"],
                      include_dirs=[numpy.get_include()]),
            Extension(name="ms_peak_picker._c.peak_index", sources=["ms_peak_picker/_c/peak_index.pyx"],
                      include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.double_vector', sources=["ms_peak_picker/_c/double_vector.pyx"]),
            Extension(name='ms_peak_picker._c.smoother', sources=["ms_peak_picker/_c/smoother.pyx"],
                      include_dirs=[numpy.get_include()]),
            OpenMPExtension(name='ms_peak_picker._c.scan_averaging',
                            sources=['ms_peak_picker/_c/scan_averaging.pyx'],
                            include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.fticr_denoising', sources=['ms_peak_picker/_c/fticr_denoising.pyx'],
                      include_dirs=[numpy.get_include()]),
            Extension(name="ms_peak_picker._c.peak_picker", sources=['ms_peak_picker/_c/peak_picker.pyx'],
                      include_dirs=[numpy.get_include()]),
        ])
    except ImportError:
        extensions = ([
            OpenMPExtension(
                name="ms_peak_picker._c.peak_statistics", sources=["ms_peak_picker/_c/peak_statistics.c"],
                include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.peak_set', sources=["ms_peak_picker/_c/peak_set.c"]),
            Extension(name='ms_peak_picker._c.fft_patterson_charge_state',
                      sources=["ms_peak_picker/_c/fft_patterson_charge_state.c"],
                      include_dirs=[numpy.get_include()]),
            Extension(name="ms_peak_picker._c.search", sources=["ms_peak_picker/_c/search.c"],
                      include_dirs=[numpy.get_include()]),
            Extension(name="ms_peak_picker._c.peak_index", sources=["ms_peak_picker/_c/peak_index.c"],
                      include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.double_vector', sources=["ms_peak_picker/_c/double_vector.c"]),
            Extension(name='ms_peak_picker._c.smoother', sources=["ms_peak_picker/_c/smoother.c"],
                      include_dirs=[numpy.get_include()]),
            OpenMPExtension(name='ms_peak_picker._c.scan_averaging',
                            sources=['ms_peak_picker/_c/scan_averaging.c'],
                            include_dirs=[numpy.get_include()]),
            Extension(name='ms_peak_picker._c.fticr_denoising', sources=['ms_peak_picker/_c/fticr_denoising.c'],
                      include_dirs=[numpy.get_include()])
        ])
    return extensions


ext_errors = (CCompilerError, DistutilsExecError, DistutilsPlatformError)
if sys.platform == 'win32':
    # 2.6's distutils.msvc9compiler can raise an IOError when failing to
    # find the compiler
    ext_errors += (IOError,)


class BuildFailed(Exception):

    def __init__(self):
        self.cause = sys.exc_info()[1]  # work around py 2/3 different syntax

    def __str__(self):
        return str(self.cause)


class ve_build_ext(build_ext):
    # This class allows C extension building to fail.

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            traceback.print_exc()
            raise BuildFailed()

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except ext_errors:
            traceback.print_exc()
            raise BuildFailed()
        except ValueError:
            # this can happen on Windows 64 bit, see Python issue 7511
            traceback.print_exc()
            if "'path'" in str(sys.exc_info()[1]):  # works with both py 2/3
                raise BuildFailed()
            raise


cmdclass = {}

cmdclass['build_ext'] = ve_build_ext


def status_msgs(*msgs):
    print('*' * 75)
    for msg in msgs:
        print(msg)
    print('*' * 75)


with open("ms_peak_picker/version.py") as version_file:
    version = None
    for line in version_file.readlines():
        if "version = " in line:
            version = line.split(" = ")[1].replace("\"", "").strip()
            print("Version is: %r" % (version,))
            break
    else:
        print("Cannot determine version")


def run_setup(include_cext=True):
    setup(
        name='ms_peak_picker',
        description='A library to pick peaks from mass spectral data',
        long_description='A library to pick peaks from mass spectral data',
        version=version,
        packages=find_packages(),
        zip_safe=False,
        install_requires=['numpy', "scipy", "six"],
        ext_modules=make_cextensions() if include_cext else None,
        cmdclass=cmdclass,
        maintainer='Joshua Klein',
        maintainer_email="jaklein@bu.edu",
        include_package_data=True,
        classifiers=[
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Topic :: Education',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Software Development :: Libraries'
        ],
        license='License :: OSI Approved :: Apache Software License')


try:
    run_setup(True)
except Exception as exc:
    if use_python_implementation:
        run_setup(False)
        status_msgs(
            "WARNING: The C extension could not be compiled, " +
            "speedups are not enabled.",
            "Plain-Python build succeeded."
        )
    else:
        raise



try:
    range = xrange
except NameError:
    range = range


from .base import Base


def ppm_error(x, y):
    return (x - y) / y


try:
    has_plot = True
    from .plot import (draw_peaklist, draw_raw)
except (RuntimeError, ImportError):
    has_plot = False


__all__ = [
    "Base", "ppm_error", "has_plot",
    "draw_raw", "draw_peaklist"
]



cimport cython
cimport numpy as np


cpdef enum PeakMode:
    profile = 1
    centroid = 2


cpdef enum PeakFit:
    quadratic = 1
    lorentzian = 2
    apex = 3


cdef class PartialPeakFitState(object):
    cdef:
        public bint set
        public double left_width
        public double right_width
        public double full_width_at_half_max
        public double signal_to_noise

    cpdef reset(self)


cdef class PeakProcessor(object):
    cdef:
        public double background_intensity
        public double _intensity_threshold
        public double _signal_to_noise_threshold

        public PartialPeakFitState partial_fit_state

        public PeakFit fit_type
        public PeakMode peak_mode

        public list peak_data

        public bint threshold_data
        public bint verbose

    cpdef double get_signal_to_noise_threshold(self)
    cpdef object set_signal_to_noise_threshold(self, double signal_to_noise_threshold)

    cpdef double get_intensity_threshold(self)
    cpdef object set_intensity_threshold(self, double intensity_threshold)


    cpdef size_t _discover_peaks(self, np.ndarray[cython.floating] mz_array, np.ndarray[cython.floating] intensity_array, double start_mz, double stop_mz)
    cpdef double find_full_width_at_half_max(self, Py_ssize_t index, np.ndarray[cython.floating, ndim=1] mz_array,
                                             np.ndarray[cython.floating, ndim=1] intensity_array, double signal_to_noise)
    cpdef double fit_peak(self, Py_ssize_t index, np.ndarray[cython.floating] mz_array, np.ndarray[cython.floating] intensity_array)
    cpdef double area(self, np.ndarray[cython.floating] mz_array, np.ndarray[cython.floating] intensity_array, double mz,
                      double full_width_at_half_max, Py_ssize_t index)